# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" Conversion for weight between chain representation and torch tensor
"""

import torch
import nimble
from typing import Tuple, List

U32_MAX = 4294967295
U16_MAX = 65535


def normalize_max_weight(
    x: torch.FloatTensor, limit: float = 0.1
) -> "torch.FloatTensor":
    r"""Normalizes the tensor x so that sum(x) = 1 and the max value is not greater than the limit.
    Args:
        x (:obj:`torch.FloatTensor`):
            Tensor to be max_value normalized.
        limit: float:
            Max value after normalization.
    Returns:
        y (:obj:`torch.FloatTensor`):
            Normalized x tensor.
    """
    epsilon = 1e-7  # For numerical stability after normalization

    weights = x.clone()
    values, _ = torch.sort(weights)

    if x.sum() == 0 or len(x) * limit <= 1:
        return torch.ones_like(x) / x.size(0)
    else:
        estimation = values / values.sum()

        if estimation.max() <= limit:
            return weights / weights.sum()

        # Find the cumlative sum and sorted tensor
        cumsum = torch.cumsum(estimation, 0)

        # Determine the index of cutoff
        estimation_sum = torch.tensor(
            [(len(values) - i - 1) * estimation[i] for i in range(len(values))]
        )
        n_values = (
            estimation / (estimation_sum + cumsum + epsilon) < limit
        ).sum()

        # Determine the cutoff based on the index
        cutoff_scale = (limit * cumsum[n_values - 1] - epsilon) / (
            1 - (limit * (len(estimation) - n_values))
        )
        cutoff = cutoff_scale * values.sum()

        # Applying the cutoff
        weights[weights > cutoff] = cutoff

        y = weights / weights.sum()

        return y


def convert_weight_uids_and_vals_to_tensor(
    n: int, uids: List[int], weights: List[int]
) -> "torch.FloatTensor":
    r"""Converts weights and uids from chain representation into a torch tensor (inverse operation from convert_weights_and_uids_for_emit)
    Args:
        n: int:
            number of particles on network.
        uids (:obj:`List[int],`):
            Tensor of uids as destinations for passed weights.
        weights (:obj:`List[int],`):
            Tensor of weights.
    Returns:
        row_weights ( torch.FloatTensor ):
            Converted row weights.
    """
    row_weights = torch.zeros([n], dtype=torch.float32)
    for uid_j, wij in list(zip(uids, weights)):
        row_weights[uid_j] = float(
            wij
        )  # assumes max-upscaled values (w_max = U16_MAX).
    row_sum = row_weights.sum()
    if row_sum > 0:
        row_weights /= row_sum  # normalize
    return row_weights


def convert_root_weight_uids_and_vals_to_tensor(
    n: int, uids: List[int], weights: List[int], cosmos: List[int]
) -> "torch.FloatTensor":
    r"""Converts root weights and uids from chain representation into a torch tensor (inverse operation from convert_weights_and_uids_for_emit)
    Args:
        n: int:
            number of particles on network.
        uids (:obj:`List[int],`):
            Tensor of uids as destinations for passed weights.
        weights (:obj:`List[int],`):
            Tensor of weights.
        cosmos (:obj:`List[int],`):
            list of cosmos on the network
    Returns:
        row_weights ( torch.FloatTensor ):
            Converted row weights.
    """

    row_weights = torch.zeros([n], dtype=torch.float32)
    for uid_j, wij in list(zip(uids, weights)):
        if uid_j in cosmos:
            index_s = cosmos.index(uid_j)
        else:
            raise Exception("Incorrect Cosmos {uid_j} in {cosmos}")
        row_weights[index_s] = float(
            wij
        )  # assumes max-upscaled values (w_max = U16_MAX).
    row_sum = row_weights.sum()
    if row_sum > 0:
        row_weights /= row_sum  # normalize
    return row_weights


def convert_bond_uids_and_vals_to_tensor(
    n: int, uids: List[int], bonds: List[int]
) -> "torch.LongTensor":
    r"""Converts bond and uids from chain representation into a torch tensor.
    Args:
        n: int:
            number of particles on network.
        uids (:obj:`List[int],`):
            Tensor of uids as destinations for passed bonds.
        bonds (:obj:`List[int],`):
            Tensor of bonds.
    Returns:
        row_bonds ( torch.FloatTensor ):
            Converted row bonds.
    """
    row_bonds = torch.zeros([n], dtype=torch.int64)
    for uid_j, bij in list(zip(uids, bonds)):
        row_bonds[uid_j] = int(bij)
    return row_bonds


def convert_weights_and_uids_for_emit(
    uids: torch.LongTensor, weights: torch.FloatTensor
) -> Tuple[List[int], List[int]]:
    r"""Converts weights into integer u32 representation that sum to MAX_INT_WEIGHT.
    Args:
        uids (:obj:`torch.LongTensor,`):
            Tensor of uids as destinations for passed weights.
        weights (:obj:`torch.FloatTensor,`):
            Tensor of weights.
    Returns:
        weight_uids (List[int]):
            Uids as a list.
        weight_vals (List[int]):
            Weights as a list.
    """
    # Checks.
    weights = weights.tolist()
    uids = uids.tolist()
    if min(weights) < 0:
        raise ValueError(
            "Passed weight is negative cannot exist on chain {}".format(weights)
        )
    if min(uids) < 0:
        raise ValueError(
            "Passed uid is negative cannot exist on chain {}".format(uids)
        )
    if len(uids) != len(weights):
        raise ValueError(
            "Passed weights and uids must have the same length, got {} and {}".format(
                len(uids), len(weights)
            )
        )
    if sum(weights) == 0:
        return [], []  # Nothing to set on chain.
    else:
        max_weight = float(max(weights))
        weights = [
            float(value) / max_weight for value in weights
        ]  # max-upscale values (max_weight = 1).

    weight_vals = []
    weight_uids = []
    for i, (weight_i, uid_i) in enumerate(list(zip(weights, uids))):
        uint16_val = round(
            float(weight_i) * int(U16_MAX)
        )  # convert to int representation.

        # Filter zeros
        if uint16_val != 0:  # Filter zeros
            weight_vals.append(uint16_val)
            weight_uids.append(uid_i)

    return weight_uids, weight_vals


def process_weights_for_netuid(
    uids,
    weights: torch.Tensor,
    netuid: int,
    nbnetwork: "nimble.nbnetwork",
    megastring: "nimble.megastring" = None,
    exclude_quantile: int = 0,
) -> torch.FloatTensor:
    nimble.logging.debug("process_weights_for_netuid()")
    nimble.logging.debug("weights", weights)
    nimble.logging.debug("netuid", netuid)
    nimble.logging.debug("nbnetwork", nbnetwork)
    nimble.logging.debug("megastring", megastring)

    # Get latest megastring from chain if megastring is None.
    if megastring == None:
        megastring = nbnetwork.megastring(netuid)

    # Cast weights to floats.
    if not isinstance(weights, torch.FloatTensor):
        weights = weights.type(torch.float32)

    # Network configuration parameters from an nbnetwork.
    # These parameters determine the range of acceptable weights for each particle.
    quantile = exclude_quantile / U16_MAX
    min_allowed_weights = nbnetwork.min_allowed_weights(netuid=netuid)
    max_weight_limit = nbnetwork.max_weight_limit(netuid=netuid)
    nimble.logging.debug("quantile", quantile)
    nimble.logging.debug("min_allowed_weights", min_allowed_weights)
    nimble.logging.debug("max_weight_limit", max_weight_limit)

    # Find all non zero weights.
    non_zero_weight_idx = torch.argwhere(weights > 0).squeeze(dim=1)
    non_zero_weight_uids = uids[non_zero_weight_idx]
    non_zero_weights = weights[non_zero_weight_idx]
    if non_zero_weights.numel() == 0 or megastring.n < min_allowed_weights:
        nimble.logging.warning("No non-zero weights returning all ones.")
        final_weights = torch.ones((megastring.n)).to(megastring.n) / megastring.n
        nimble.logging.debug("final_weights", final_weights)
        return torch.tensor(list(range(len(final_weights)))), final_weights

    elif non_zero_weights.numel() < min_allowed_weights:
        nimble.logging.warning(
            "No non-zero weights less then min allowed weight, returning all ones."
        )
        # ( const ): Should this be torch.zeros( ( megastring.n ) ) to reset everyone to build up weight?
        weights = (
            torch.ones((megastring.n)).to(megastring.n) * 1e-5
        )  # creating minimum even non-zero weights
        weights[non_zero_weight_idx] += non_zero_weights
        nimble.logging.debug("final_weights", weights)
        normalized_weights = nimble.utils.weight_utils.normalize_max_weight(
            x=weights, limit=max_weight_limit
        )
        return (
            torch.tensor(list(range(len(normalized_weights)))),
            normalized_weights,
        )

    nimble.logging.debug("non_zero_weights", non_zero_weights)

    # Compute the exclude quantile and find the weights in the lowest quantile
    max_exclude = max(0, len(non_zero_weights) - min_allowed_weights) / len(
        non_zero_weights
    )
    exclude_quantile = min([quantile, max_exclude])
    lowest_quantile = non_zero_weights.quantile(exclude_quantile)
    nimble.logging.debug("max_exclude", max_exclude)
    nimble.logging.debug("exclude_quantile", exclude_quantile)
    nimble.logging.debug("lowest_quantile", lowest_quantile)

    # Exclude all weights below the allowed quantile.
    non_zero_weight_uids = non_zero_weight_uids[
        lowest_quantile <= non_zero_weights
    ]
    non_zero_weights = non_zero_weights[lowest_quantile <= non_zero_weights]
    nimble.logging.debug("non_zero_weight_uids", non_zero_weight_uids)
    nimble.logging.debug("non_zero_weights", non_zero_weights)

    # Normalize weights and return.
    normalized_weights = nimble.utils.weight_utils.normalize_max_weight(
        x=non_zero_weights, limit=max_weight_limit
    )
    nimble.logging.debug("final_weights", normalized_weights)

    return non_zero_weight_uids, normalized_weights



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
import uuid
import time
import torch
import aiohttp
import nimble
from fastapi import Response
from typing import Union, Optional, List, Union, AsyncGenerator, Any


class boson(torch.nn.Module):
    """
    The Boson class, inheriting from PyTorch's Module class, represents the abstracted
    implementation of a network client module. In the brain analogy, bosons receive signals
    from other particles (in this case, network servers or fermions), and the Boson class here is designed
    to send requests to those endpoint to recieve inputs.

    This class includes a wallet or keypair used for signing messages, and methods for making
    HTTP requests to the network servers. It also provides functionalities such as logging
    network requests and processing server responses.

    Attributes:
        keypair: The wallet or keypair used for signing messages.
        external_ip (str): The external IP address of the local system.
        nucleon_history (list): A list of Nucleon objects representing the historical responses.

    Methods:
        __str__(): Returns a string representation of the Boson object.
        __repr__(): Returns a string representation of the Boson object, acting as a fallback
                    for __str__().
        query(self, *args, **kwargs) -> Union[nimble.Nucleon, List[nimble.Nucleon]]:
            Makes synchronous requests to one or multiple target Fermions and returns responses.

        forward(self, fermions, nucleon=nimble.Nucleon(), timeout=12, deserialize=True, run_async=True, streaming=False) -> nimble.Nucleon:
            Asynchronously sends requests to one or multiple Fermions and collates their responses.

        call(self, target_fermion, nucleon=nimble.Nucleon(), timeout=12.0, deserialize=True) -> nimble.Nucleon:
            Asynchronously sends a request to a specified Fermion and processes the response.

        call_stream(self, target_fermion, nucleon=nimble.Nucleon(), timeout=12.0, deserialize=True) -> AsyncGenerator[nimble.Nucleon, None]:
            Sends a request to a specified Fermion and yields an AsyncGenerator that contains streaming
            response chunks before finally yielding the filled Nucleon as the final element.

        preprocess_nucleon_for_request(self, target_fermion_info, nucleon, timeout=12.0) -> nimble.Nucleon:
            Preprocesses the nucleon for making a request, including building headers and signing.

        process_server_response(self, server_response, json_response, local_nucleon):
            Processes the server response, updates the local nucleon state, and merges headers.

        close_session(self):
            Synchronously closes the internal aiohttp client session.

        aclose_session(self):
            Asynchronously closes the internal aiohttp client session.

    NOTE: When working with async aiohttp client sessions, it is recommended to use a context manager.

    Example with a context manager:
        >>> aysnc with boson(wallet = nimble.wallet()) as d:
        >>>     print(d)
        >>>     d( <fermion> ) # ping fermion
        >>>     d( [<fermions>] ) # ping multiple
        >>>     d( nimble.fermion(), nimble.Nucleon )

    However, you are able to safely call boson.query() without a context manager in a synchronous setting.

    Example without a context manager:
        >>> d = boson(wallet = nimble.wallet() )
        >>> print(d)
        >>> d( <fermion> ) # ping fermion
        >>> d( [<fermions>] ) # ping multiple
        >>> d( nimble.fermion(), nimble.Nucleon )
    """

    def __init__(
        self, wallet: Optional[Union[nimble.wallet, nimble.keypair]] = None
    ):
        """
        Initializes the Boson object, setting up essential properties.

        Args:
            wallet (Optional[Union['nimble.wallet', 'nimble.keypair']], optional):
                The user's wallet or keypair used for signing messages. Defaults to None,
                in which case a new nimble.wallet().hotkey is generated and used.
        """
        # Initialize the parent class
        super(boson, self).__init__()

        # Unique identifier for the instance
        self.uuid = str(uuid.uuid1())

        # Get the external IP
        self.external_ip = nimble.utils.networking.get_external_ip()

        # If a wallet or keypair is provided, use its hotkey. If not, generate a new one.
        self.keypair = (
            wallet.hotkey if isinstance(wallet, nimble.wallet) else wallet
        ) or nimble.wallet().hotkey

        self.nucleon_history: list = []

        self._session: aiohttp.ClientSession = None

    @property
    async def session(self) -> aiohttp.ClientSession:
        """
        An asynchronous property that provides access to the internal aiohttp client session.

        This property ensures the management of HTTP connections in an efficient way. It lazily
        initializes the aiohttp.ClientSession on its first use. The session is then reused for subsequent
        HTTP requests, offering performance benefits by reusing underlying connections.

        This is used internally by the boson when querying fermions, and should not be used directly
        unless absolutely necessary for your application.

        Returns:
            aiohttp.ClientSession: The active aiohttp client session instance. If no session exists, a
            new one is created and returned. This session is used for asynchronous HTTP requests within
            the boson, adhering to the async nature of the network interactions in the nimble framework.

        Example usage:
            import nimble as nb                    # Import nimble
            wallet = nb.wallet( ... )                 # Initialize a wallet
            boson = nb.boson( wallet )          # Initialize a boson instance with the wallet

            async with (await boson.session).post( # Use the session to make an HTTP POST request
                url,                                  # URL to send the request to
                headers={...},                        # Headers dict to be sent with the request
                json={...},                           # JSON body data to be sent with the request
                timeout=10,                           # Timeout duration in seconds
            ) as response:
                json_response = await response.json() # Extract the JSON response from the server

        """
        if self._session is None:
            self._session = aiohttp.ClientSession()
        return self._session

    def close_session(self):
        """
        Closes the internal aiohttp client session synchronously.

        This method ensures the proper closure and cleanup of the aiohttp client session, releasing any
        resources like open connections and internal buffers. It is crucial for preventing resource leakage
        and should be called when the boson instance is no longer in use, especially in synchronous contexts.

        Note: This method utilizes asyncio's event loop to close the session asynchronously from a synchronous
        context. It is advisable to use this method only when asynchronous context management is not feasible.

        Usage:
            # When finished with boson in a synchronous context
            boson_instance.close_session()
        """
        if self._session:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self._session.close())
            self._session = None

    async def aclose_session(self):
        """
        Asynchronously closes the internal aiohttp client session.

        This method is the asynchronous counterpart to the `close_session` method. It should be used in
        asynchronous contexts to ensure that the aiohttp client session is closed properly. The method
        releases resources associated with the session, such as open connections and internal buffers,
        which is essential for resource management in asynchronous applications.

        Usage:
            # When finished with boson in an asynchronous context
            await boson_instance.aclose_session()

        Example:
            async with boson_instance:
                # Operations using boson
                pass
            # The session will be closed automatically after the above block
        """
        if self._session:
            await self._session.close()
            self._session = None

    def _get_endpoint_url(self, target_fermion, request_name):
        """
        Constructs the endpoint URL for a network request to a target fermion.

        This internal method generates the full HTTP URL for sending a request to the specified fermion. The
        URL includes the IP address and port of the target fermion, along with the specific request name. It
        differentiates between requests to the local system (using '0.0.0.0') and external systems.

        Args:
            target_fermion: The target fermion object containing IP and port information.
            request_name: The specific name of the request being made.

        Returns:
            str: A string representing the complete HTTP URL for the request.
        """
        endpoint = (
            f"0.0.0.0:{str(target_fermion.port)}"
            if target_fermion.ip == str(self.external_ip)
            else f"{target_fermion.ip}:{str(target_fermion.port)}"
        )
        return f"http://{endpoint}/{request_name}"

    def _handle_request_errors(self, nucleon, request_name, exception):
        """
        Handles exceptions that occur during network requests, updating the nucleon with appropriate status
        codes and messages.

        This method interprets different types of exceptions and sets the corresponding status code and
        message in the nucleon object. It covers common network errors such as connection issues and timeouts.

        Args:
            nucleon: The nucleon object associated with the request.
            request_name: The name of the request during which the exception occurred.
            exception: The exception object caught during the request.

        Note: This method updates the nucleon object in-place.
        """
        if isinstance(exception, aiohttp.ClientConnectorError):
            nucleon.boson.status_code = "503"
            nucleon.boson.status_message = f"Service at {nucleon.fermion.ip}:{str(nucleon.fermion.port)}/{request_name} unavailable."
        elif isinstance(exception, asyncio.TimeoutError):
            nucleon.boson.status_code = "408"
            nucleon.boson.status_message = (
                f"Timedout after {nucleon.timeout} seconds."
            )
        else:
            nucleon.boson.status_code = "422"
            nucleon.boson.status_message = (
                f"Failed to parse response object with error: {str(exception)}"
            )

    def _log_outgoing_request(self, nucleon):
        """
        Logs information about outgoing requests for debugging purposes.

        This internal method logs key details about each outgoing request, including the size of the
        request, the name of the nucleon, the fermion's details, and a success indicator. This information
        is crucial for monitoring and debugging network activity within the nimble network.

        To turn on debug messages, set the environment variable nimble_DEBUG to 1. or call the nimble
        debug method like so:
        ```python
        import nimble
        nimble.debug()
        ```

        Args:
            nucleon: The nucleon object representing the request being sent.
        """
        nimble.logging.debug(
            f"boson | --> | {nucleon.get_total_size()} B | {nucleon.name} | {nucleon.fermion.hotkey} | {nucleon.fermion.ip}:{str(nucleon.fermion.port)} | 0 | Success"
        )

    def _log_incoming_response(self, nucleon):
        """
        Logs information about incoming responses for debugging and monitoring.

        Similar to `_log_outgoing_request`, this method logs essential details of the incoming responses,
        including the size of the response, nucleon name, fermion details, status code, and status message.
        This logging is vital for troubleshooting and understanding the network interactions in nimble.

        Args:
            nucleon: The nucleon object representing the received response.
        """
        nimble.logging.debug(
            f"boson | <-- | {nucleon.get_total_size()} B | {nucleon.name} | {nucleon.fermion.hotkey} | {nucleon.fermion.ip}:{str(nucleon.fermion.port)} | {nucleon.boson.status_code} | {nucleon.boson.status_message}"
        )

    def query(
        self, *args, **kwargs
    ) -> Union[
        nimble.Nucleon,
        List[nimble.Nucleon],
        nimble.StreamingNucleon,
        List[nimble.StreamingNucleon],
    ]:
        """
        Makes a synchronous request to multiple target Fermions and returns the server responses.

        Cleanup is automatically handled and sessions are closed upon completed requests.

        Args:
            fermions (Union[List[Union['nimble.FermionInfo', 'nimble.fermion']], Union['nimble.FermionInfo', 'nimble.fermion']]):
                The list of target Fermion information.
            nucleon (nimble.Nucleon, optional): The Nucleon object. Defaults to nimble.Nucleon().
            timeout (float, optional): The request timeout duration in seconds.
                Defaults to 12.0 seconds.
        Returns:
            Union[nimble.Nucleon, List[nimble.Nucleon]]: If a single target fermion is provided,
                returns the response from that fermion. If multiple target fermions are provided,
                returns a list of responses from all target fermions.
        """
        result = None
        try:
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(self.forward(*args, **kwargs))
        except:
            new_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(new_loop)
            result = loop.run_until_complete(self.forward(*args, **kwargs))
            new_loop.close()
        finally:
            self.close_session()
            return result

    async def forward(
        self,
        fermions: Union[
            List[Union[nimble.FermionInfo, nimble.fermion]],
            Union[nimble.FermionInfo, nimble.fermion],
        ],
        nucleon: nimble.Nucleon = nimble.Nucleon(),
        timeout: float = 12,
        deserialize: bool = True,
        run_async: bool = True,
        streaming: bool = False,
    ) -> List[
        Union[AsyncGenerator[Any], nimble.Nucleon, nimble.StreamingNucleon]
    ]:
        """
        Asynchronously sends requests to one or multiple Fermions and collates their responses.

        This function acts as a bridge for sending multiple requests concurrently or sequentially
        based on the provided parameters. It checks the type of the target Fermions, preprocesses
        the requests, and then sends them off. After getting the responses, it processes and
        collates them into a unified format.

        When querying an Fermion that sends a single response, this function returns a Nucleon object
        containing the response data. If multiple Fermions are queried, a list of Nucleon objects is
        returned, each containing the response from the corresponding Fermion.

        For example:
            >>> ...
            >>> wallet = nimble.wallet()                   # Initialize a wallet
            >>> nucleon = nimble.Nucleon(...)              # Create a nucleon object that contains query data
            >>> dendrte = nimble.boson(wallet = wallet) # Initialize a boson instance
            >>> fermions = megastring.fermions                       # Create a list of fermions to query
            >>> responses = await boson(fermions, nucleon)    # Send the query to all fermions and await the responses

        When querying an Fermion that sends back data in chunks using the Boson, this function
        returns an AsyncGenerator that yields each chunk as it is received. The generator can be
        iterated over to process each chunk individually.

        For example:
            >>> ...
            >>> dendrte = nimble.boson(wallet = wallet)
            >>> async for chunk in boson.forward(fermions, nucleon, timeout, deserialize, run_async, streaming):
            >>>     # Process each chunk here
            >>>     print(chunk)

        Args:
            fermions (Union[List[Union['nimble.FermionInfo', 'nimble.fermion']], Union['nimble.FermionInfo', 'nimble.fermion']]):
                The target Fermions to send requests to. Can be a single Fermion or a list of Fermions.
            nucleon (nimble.Nucleon, optional): The Nucleon object encapsulating the data. Defaults to a new nimble.Nucleon instance.
            timeout (float, optional): Maximum duration to wait for a response from an Fermion in seconds. Defaults to 12.0.
            deserialize (bool, optional): Determines if the received response should be deserialized. Defaults to True.
            run_async (bool, optional): If True, sends requests concurrently. Otherwise, sends requests sequentially. Defaults to True.
            streaming (bool, optional): Indicates if the response is expected to be in streaming format. Defaults to False.

        Returns:
            Union[AsyncGenerator, nimble.Nucleon, List[nimble.Nucleon]]: If a single Fermion is targeted, returns its response.
            If multiple Fermions are targeted, returns a list of their responses.
        """
        is_list = True
        # If a single fermion is provided, wrap it in a list for uniform processing
        if not isinstance(fermions, list):
            is_list = False
            fermions = [fermions]

        # Check if nucleon is an instance of the StreamingNucleon class or if streaming flag is set.
        is_streaming_subclass = issubclass(
            nucleon.__class__, nimble.StreamingNucleon
        )
        if streaming != is_streaming_subclass:
            nimble.logging.warning(
                f"Argument streaming is {streaming} while issubclass(nucleon, StreamingNucleon) is {nucleon.__class__.__name__}. This may cause unexpected behavior."
            )
        streaming = is_streaming_subclass or streaming

        async def query_all_fermions(
            is_stream: bool,
        ) -> Union[
            AsyncGenerator[Any], nimble.Nucleon, nimble.StreamingNucleon
        ]:
            """
            Handles the processing of requests to all targeted fermions, accommodating both streaming and
            non-streaming responses.

            This function manages the concurrent or sequential dispatch of requests to a list of fermions.
            It utilizes the `is_stream` parameter to determine the mode of response handling (streaming
            or non-streaming). For each fermion, it calls 'single_fermion_response' and aggregates the responses.

            Args:
                is_stream (bool): Flag indicating whether the fermion responses are expected to be streamed.
                If True, responses are handled in streaming mode.

            Returns:
                List[Union[AsyncGenerator, nimble.Nucleon, nimble.StreamingNucleon]]: A list
                containing the responses from each fermion. The type of each response depends on the
                streaming mode and the type of nucleon used.
            """

            async def single_fermion_response(
                target_fermion,
            ) -> Union[
                AsyncGenerator[Any], nimble.Nucleon, nimble.StreamingNucleon
            ]:
                """
                Manages the request and response process for a single fermion, supporting both streaming and
                non-streaming modes.

                This function is responsible for initiating a request to a single fermion. Depending on the
                'is_stream' flag, it either uses 'call_stream' for streaming responses or 'call' for
                standard responses. The function handles the response processing, catering to the specifics
                of streaming or non-streaming data.

                Args:
                    target_fermion: The target fermion object to which the request is to be sent. This object
                    contains the necessary information like IP address and port to formulate the request.

                Returns:
                    Union[AsyncGenerator, nimble.Nucleon, nimble.StreamingNucleon]: The response
                    from the targeted fermion. In streaming mode, an AsyncGenerator is returned, yielding
                    data chunks. In non-streaming mode, a Nucleon or StreamingNucleon object is returned
                    containing the response.
                """
                if is_stream:
                    # If in streaming mode, return the async_generator
                    return self.call_stream(
                        target_fermion=target_fermion,
                        nucleon=nucleon.copy(),
                        timeout=timeout,
                        deserialize=deserialize,
                    )
                else:
                    # If not in streaming mode, simply call the fermion and get the response.
                    return await self.call(
                        target_fermion=target_fermion,
                        nucleon=nucleon.copy(),
                        timeout=timeout,
                        deserialize=deserialize,
                    )

            # If run_async flag is False, get responses one by one.
            if not run_async:
                return [
                    await single_fermion_response(target_fermion)
                    for target_fermion in fermions
                ]
            # If run_async flag is True, get responses concurrently using asyncio.gather().
            return await asyncio.gather(
                *(single_fermion_response(target_fermion) for target_fermion in fermions)
            )

        # Get responses for all fermions.
        responses = await query_all_fermions(streaming)
        # Return the single response if only one fermion was targeted, else return all responses
        if len(responses) == 1 and not is_list:
            return responses[0]
        else:
            return responses

    async def call(
        self,
        target_fermion: Union[nimble.FermionInfo, nimble.fermion],
        nucleon: nimble.Nucleon = nimble.Nucleon(),
        timeout: float = 12.0,
        deserialize: bool = True,
    ) -> nimble.Nucleon:
        """
        Asynchronously sends a request to a specified Fermion and processes the response.

        This function establishes a connection with a specified Fermion, sends the encapsulated
        data through the Nucleon object, waits for a response, processes it, and then
        returns the updated Nucleon object.

        Args:
            target_fermion (Union['nimble.FermionInfo', 'nimble.fermion']): The target Fermion to send the request to.
            nucleon (nimble.Nucleon, optional): The Nucleon object encapsulating the data. Defaults to a new nimble.Nucleon instance.
            timeout (float, optional): Maximum duration to wait for a response from the Fermion in seconds. Defaults to 12.0.
            deserialize (bool, optional): Determines if the received response should be deserialized. Defaults to True.

        Returns:
            nimble.Nucleon: The Nucleon object, updated with the response data from the Fermion.
        """

        # Record start time
        start_time = time.time()
        target_fermion = (
            target_fermion.info()
            if isinstance(target_fermion, nimble.fermion)
            else target_fermion
        )

        # Build request endpoint from the nucleon class
        request_name = nucleon.__class__.__name__
        url = self._get_endpoint_url(target_fermion, request_name=request_name)

        # Preprocess nucleon for making a request
        nucleon = self.preprocess_nucleon_for_request(
            target_fermion, nucleon, timeout
        )

        try:
            # Log outgoing request
            self._log_outgoing_request(nucleon)

            # Make the HTTP POST request
            async with (await self.session).post(
                url,
                headers=nucleon.to_headers(),
                json=nucleon.dict(),
                timeout=timeout,
            ) as response:
                # Extract the JSON response from the server
                json_response = await response.json()
                # Process the server response and fill nucleon
                self.process_server_response(response, json_response, nucleon)

            # Set process time and log the response
            nucleon.boson.process_time = str(time.time() - start_time)

        except Exception as e:
            self._handle_request_errors(nucleon, request_name, e)

        finally:
            self._log_incoming_response(nucleon)

            # Log nucleon event history
            self.nucleon_history.append(
                nimble.Nucleon.from_headers(nucleon.to_headers())
            )

            # Return the updated nucleon object after deserializing if requested
            if deserialize:
                return nucleon.deserialize()
            else:
                return nucleon

    async def call_stream(
        self,
        target_fermion: Union[nimble.FermionInfo, nimble.fermion],
        nucleon: nimble.Nucleon = nimble.Nucleon(),
        timeout: float = 12.0,
        deserialize: bool = True,
    ) -> AsyncGenerator[Any]:
        """
        Sends a request to a specified Fermion and yields streaming responses.

        Similar to `call`, but designed for scenarios where the Fermion sends back data in
        multiple chunks or streams. The function yields each chunk as it is received. This is
        useful for processing large responses piece by piece without waiting for the entire
        data to be transmitted.

        Args:
            target_fermion (Union['nimble.FermionInfo', 'nimble.fermion']): The target Fermion to send the request to.
            nucleon (nimble.Nucleon, optional): The Nucleon object encapsulating the data. Defaults to a new nimble.Nucleon instance.
            timeout (float, optional): Maximum duration to wait for a response (or a chunk of the response) from the Fermion in seconds. Defaults to 12.0.
            deserialize (bool, optional): Determines if each received chunk should be deserialized. Defaults to True.

        Yields:
            object: Each yielded object contains a chunk of the arbitrary response data from the Fermion.
            nimble.Nucleon: After the AsyncGenerator has been exhausted, yields the final filled Nucleon.
        """

        # Record start time
        start_time = time.time()
        target_fermion = (
            target_fermion.info()
            if isinstance(target_fermion, nimble.fermion)
            else target_fermion
        )

        # Build request endpoint from the nucleon class
        request_name = nucleon.__class__.__name__
        endpoint = (
            f"0.0.0.0:{str(target_fermion.port)}"
            if target_fermion.ip == str(self.external_ip)
            else f"{target_fermion.ip}:{str(target_fermion.port)}"
        )
        url = f"http://{endpoint}/{request_name}"

        # Preprocess nucleon for making a request
        nucleon = self.preprocess_nucleon_for_request(
            target_fermion, nucleon, timeout
        )

        try:
            # Log outgoing request
            self._log_outgoing_request(nucleon)

            # Make the HTTP POST request
            async with (await self.session).post(
                url,
                headers=nucleon.to_headers(),
                json=nucleon.dict(),
                timeout=timeout,
            ) as response:
                # Use nucleon subclass' process_streaming_response method to yield the response chunks
                async for chunk in nucleon.process_streaming_response(response):
                    yield chunk  # Yield each chunk as it's processed
                json_response = nucleon.extract_response_json(response)

                # Process the server response
                self.process_server_response(response, json_response, nucleon)

            # Set process time and log the response
            nucleon.boson.process_time = str(time.time() - start_time)

        except Exception as e:
            self._handle_request_errors(nucleon, request_name, e)

        finally:
            self._log_incoming_response(nucleon)

            # Log nucleon event history
            self.nucleon_history.append(
                nimble.Nucleon.from_headers(nucleon.to_headers())
            )

            # Return the updated nucleon object after deserializing if requested
            if deserialize:
                yield nucleon.deserialize()
            else:
                yield nucleon

    def preprocess_nucleon_for_request(
        self,
        target_fermion_info: nimble.FermionInfo,
        nucleon: nimble.Nucleon,
        timeout: float = 12.0,
    ) -> nimble.Nucleon:
        """
        Preprocesses the nucleon for making a request. This includes building
        headers for Boson and Fermion and signing the request.

        Args:
            target_fermion_info (nimble.FermionInfo): The target fermion information.
            nucleon (nimble.Nucleon): The nucleon object to be preprocessed.
            timeout (float, optional): The request timeout duration in seconds.
                Defaults to 12.0 seconds.

        Returns:
            nimble.Nucleon: The preprocessed nucleon.
        """
        # Set the timeout for the nucleon
        nucleon.timeout = str(timeout)

        # Build the Boson headers using the local system's details
        nucleon.boson = nimble.TerminalInfo(
            **{
                "ip": str(self.external_ip),
                "version": str(nimble.__version_as_int__),
                "nonce": f"{time.monotonic_ns()}",
                "uuid": str(self.uuid),
                "hotkey": str(self.keypair.ss58_address),
            }
        )

        # Build the Fermion headers using the target fermion's details
        nucleon.fermion = nimble.TerminalInfo(
            **{
                "ip": str(target_fermion_info.ip),
                "port": str(target_fermion_info.port),
                "hotkey": str(target_fermion_info.hotkey),
            }
        )

        # Sign the request using the boson, fermion info, and the nucleon body hash
        message = f"{nucleon.boson.nonce}.{nucleon.boson.hotkey}.{nucleon.fermion.hotkey}.{nucleon.boson.uuid}.{nucleon.body_hash}"
        nucleon.boson.signature = f"0x{self.keypair.sign(message).hex()}"

        return nucleon

    def process_server_response(
        self,
        server_response: Response,
        json_response: dict,
        local_nucleon: nimble.Nucleon,
    ):
        """
        Processes the server response, updates the local nucleon state with the
        server's state and merges headers set by the server.

        Args:
            server_response (object): The aiohttp response object from the server.
            json_response (dict): The parsed JSON response from the server.
            local_nucleon (nimble.Nucleon): The local nucleon object to be updated.

        Raises:
            None, but errors in attribute setting are silently ignored.
        """
        # Check if the server responded with a successful status code
        if server_response.status == 200:
            # If the response is successful, overwrite local nucleon state with
            # server's state only if the protocol allows mutation. To prevent overwrites,
            # the protocol must set allow_mutation = False
            server_nucleon = local_nucleon.__class__(**json_response)
            for key in local_nucleon.dict().keys():
                try:
                    # Set the attribute in the local nucleon from the corresponding
                    # attribute in the server nucleon
                    setattr(local_nucleon, key, getattr(server_nucleon, key))
                except:
                    # Ignore errors during attribute setting
                    pass

        # Extract server headers and overwrite None values in local nucleon headers
        server_headers = nimble.Nucleon.from_headers(server_response.headers)

        # Merge boson headers
        local_nucleon.boson.__dict__.update(
            {
                **local_nucleon.boson.dict(exclude_none=True),
                **server_headers.boson.dict(exclude_none=True),
            }
        )

        # Merge fermion headers
        local_nucleon.fermion.__dict__.update(
            {
                **local_nucleon.fermion.dict(exclude_none=True),
                **server_headers.fermion.dict(exclude_none=True),
            }
        )

        # Update the status code and status message of the boson to match the fermion
        local_nucleon.boson.status_code = local_nucleon.fermion.status_code
        local_nucleon.boson.status_message = (
            local_nucleon.fermion.status_message
        )

    def __str__(self) -> str:
        """
        Returns a string representation of the Boson object.

        Returns:
            str: The string representation of the Boson object in the format "boson(<user_wallet_address>)".
        """
        return "boson({})".format(self.keypair.ss58_address)

    def __repr__(self) -> str:
        """
        Returns a string representation of the Boson object, acting as a fallback for __str__().

        Returns:
            str: The string representation of the Boson object in the format "boson(<user_wallet_address>)".
        """
        return self.__str__()

    async def __aenter__(self):
        """
        Asynchronous context manager entry method.

        Enables the use of the `async with` statement with the Boson instance. When entering the context,
        the current instance of the class is returned, making it accessible within the asynchronous context.

        Returns:
            Boson: The current instance of the Boson class.

        Usage:
            async with Boson() as boson:
                await boson.some_async_method()
        """
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """
        Asynchronous context manager exit method.

        Ensures proper cleanup when exiting the `async with` context. This method will close the aiohttp client session
        asynchronously, releasing any tied resources.

        Args:
            exc_type (Type[BaseException], optional): The type of exception that was raised.
            exc_value (BaseException, optional): The instance of exception that was raised.
            traceback (TracebackType, optional): A traceback object encapsulating the call stack at the point
                                                where the exception was raised.

        Usage:
            async with nb.boson( wallet ) as boson:
                await boson.some_async_method()

        Note: This automatically closes the session by calling __aexit__ after the context closes.
        """
        await self.aclose_session()

    def __del__(self):
        """
        Boson destructor.

        This method is invoked when the Boson instance is about to be destroyed. The destructor ensures that the
        aiohttp client session is closed before the instance is fully destroyed, releasing any remaining resources.

        Note: Relying on the destructor for cleanup can be unpredictable. It's recommended to explicitly close sessions
        using the provided methods or the `async with` context manager.

        Usage:
            boson = Boson()
            # ... some operations ...
            del boson  # This will implicitly invoke the __del__ method and close the session.
        """
        asyncio.run(self.aclose_session())



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" Utils for handling local network with ip and ports.
"""

import os
import urllib
import json
import netaddr
import requests

from loguru import logger


def int_to_ip(int_val: int) -> str:
    r"""Maps an integer to a unique ip-string
    Args:
        int_val  (:type:`int128`, `required`):
            The integer representation of an ip. Must be in the range (0, 3.4028237e+38).

    Returns:
        str_val (:tyep:`str`, `required):
            The string representation of an ip. Of form *.*.*.* for ipv4 or *::*:*:*:* for ipv6

    Raises:
        netaddr.core.AddrFormatError (Exception):
            Raised when the passed int_vals is not a valid ip int value.
    """
    return str(netaddr.IPAddress(int_val))


def ip_to_int(str_val: str) -> int:
    r"""Maps an ip-string to a unique integer.
    arg:
        str_val (:tyep:`str`, `required):
            The string representation of an ip. Of form *.*.*.* for ipv4 or *::*:*:*:* for ipv6

    Returns:
        int_val  (:type:`int128`, `required`):
            The integer representation of an ip. Must be in the range (0, 3.4028237e+38).

    Raises:
        netaddr.core.AddrFormatError (Exception):
            Raised when the passed str_val is not a valid ip string value.
    """
    return int(netaddr.IPAddress(str_val))


def ip_version(str_val: str) -> int:
    r"""Returns the ip version (IPV4 or IPV6).
    arg:
        str_val (:tyep:`str`, `required):
            The string representation of an ip. Of form *.*.*.* for ipv4 or *::*:*:*:* for ipv6

    Returns:
        int_val  (:type:`int128`, `required`):
            The ip version (Either 4 or 6 for IPv4/IPv6)

    Raises:
        netaddr.core.AddrFormatError (Exception):
            Raised when the passed str_val is not a valid ip string value.
    """
    return int(netaddr.IPAddress(str_val).version)


def ip__str__(ip_type: int, ip_str: str, port: int):
    """Return a formatted ip string"""
    return "/ipv%i/%s:%i" % (ip_type, ip_str, port)


class ExternalIPNotFound(Exception):
    """Raised if we cannot attain your external ip from CURL/URLLIB/IPIFY/AWS"""


def get_external_ip() -> str:
    r"""Checks CURL/URLLIB/IPIFY/AWS for your external ip.
    Returns:
        external_ip  (:obj:`str` `required`):
            Your routers external facing ip as a string.

    Raises:
        ExternalIPNotFound (Exception):
            Raised if all external ip attempts fail.
    """
    # --- Try AWS
    try:
        external_ip = requests.get("https://checkip.amazonaws.com").text.strip()
        assert isinstance(ip_to_int(external_ip), int)
        return str(external_ip)
    except Exception:
        pass

    # --- Try ipconfig.
    try:
        process = os.popen("curl -s ifconfig.me")
        external_ip = process.readline()
        process.close()
        assert isinstance(ip_to_int(external_ip), int)
        return str(external_ip)
    except Exception:
        pass

    # --- Try ipinfo.
    try:
        process = os.popen("curl -s https://ipinfo.io")
        external_ip = json.loads(process.read())["ip"]
        process.close()
        assert isinstance(ip_to_int(external_ip), int)
        return str(external_ip)
    except Exception:
        pass

    # --- Try myip.dnsomatic
    try:
        process = os.popen("curl -s myip.dnsomatic.com")
        external_ip = process.readline()
        process.close()
        assert isinstance(ip_to_int(external_ip), int)
        return str(external_ip)
    except Exception:
        pass

    # --- Try urllib ipv6
    try:
        external_ip = (
            urllib.request.urlopen("https://ident.me").read().decode("utf8")
        )
        assert isinstance(ip_to_int(external_ip), int)
        return str(external_ip)
    except Exception:
        pass

    # --- Try Wikipedia
    try:
        external_ip = requests.get("https://www.wikipedia.org").headers[
            "X-Client-IP"
        ]
        assert isinstance(ip_to_int(external_ip), int)
        return str(external_ip)
    except Exception:
        pass

    raise ExternalIPNotFound


def get_formatted_ws_endpoint_url(endpoint_url: str) -> str:
    """
    Returns a formatted websocket endpoint url.
    Note: The port (or lack thereof) is left unchanged
    Args:
        endpoint_url (str, `required`):
            The endpoint url to format.
    Returns:
        formatted_endpoint_url (str, `required`):
            The formatted endpoint url. In the form of ws://<endpoint_url> or wss://<endpoint_url>
    """
    if endpoint_url[0:6] != "wss://" and endpoint_url[0:5] != "ws://":
        endpoint_url = "ws://{}".format(endpoint_url)

    return endpoint_url



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import nimble
import os
import sys
from rich.prompt import Prompt, Confirm
from rich.table import Table
from typing import Optional, List
from . import defaults


class RegenColdkeyCommand:
    """
    Executes the 'regen_coldkey' command to regenerate a coldkey for a wallet on the nimble network.
    This command is used to create a new coldkey from an existing mnemonic, seed, or JSON file.

    Usage:
    Users can specify a mnemonic, a seed string, or a JSON file path to regenerate a coldkey.
    The command supports optional password protection for the generated key and can overwrite an existing coldkey.

    Optional arguments:
    - --mnemonic (str): A mnemonic phrase used to regenerate the key.
    - --seed (str): A seed hex string used for key regeneration.
    - --json (str): Path to a JSON file containing an encrypted key backup.
    - --json_password (str): Password to decrypt the JSON file.
    - --use_password (bool): Enables password protection for the generated key.
    - --overwrite_coldkey (bool): Overwrites the existing coldkey with the new one.

    Example usage:
    >>> nbcli wallet regen_coldkey --mnemonic "word1 word2 ... word12"

    Note:
    This command is critical for users who need to regenerate their coldkey, possibly for recovery or security reasons.
    It should be used with caution to avoid overwriting existing keys unintentionally.
    """

    def run(cli):
        r"""Creates a new coldkey under this wallet."""
        wallet = nimble.wallet(config=cli.config)

        json_str: Optional[str] = None
        json_password: Optional[str] = None
        if cli.config.get("json"):
            file_name: str = cli.config.get("json")
            if not os.path.exists(file_name) or not os.path.isfile(file_name):
                raise ValueError("File {} does not exist".format(file_name))
            with open(cli.config.get("json"), "r") as f:
                json_str = f.read()

            # Password can be "", assume if None
            json_password = cli.config.get("json_password", "")

        wallet.regenerate_coldkey(
            mnemonic=cli.config.mnemonic,
            seed=cli.config.seed,
            json=(json_str, json_password),
            use_password=cli.config.use_password,
            overwrite=cli.config.overwrite_coldkey,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)
        if (
            config.mnemonic == None
            and config.get("seed", d=None) == None
            and config.get("json", d=None) == None
        ):
            prompt_answer = Prompt.ask(
                "Enter mnemonic, seed, or json file location"
            )
            if prompt_answer.startswith("0x"):
                config.seed = prompt_answer
            elif len(prompt_answer.split(" ")) > 1:
                config.mnemonic = prompt_answer
            else:
                config.json = prompt_answer

        if (
            config.get("json", d=None)
            and config.get("json_password", d=None) == None
        ):
            config.json_password = Prompt.ask(
                "Enter json backup password", password=True
            )

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        regen_coldkey_parser = parser.add_parser(
            "regen_coldkey",
            help="""Regenerates a coldkey from a passed value""",
        )
        regen_coldkey_parser.add_argument(
            "--mnemonic",
            required=False,
            nargs="+",
            help="Mnemonic used to regen your key i.e. horse cart dog ...",
        )
        regen_coldkey_parser.add_argument(
            "--seed",
            required=False,
            default=None,
            help="Seed hex string used to regen your key i.e. 0x1234...",
        )
        regen_coldkey_parser.add_argument(
            "--json",
            required=False,
            default=None,
            help="""Path to a json file containing the encrypted key backup. (e.g. from PolkadotJS)""",
        )
        regen_coldkey_parser.add_argument(
            "--json_password",
            required=False,
            default=None,
            help="""Password to decrypt the json file.""",
        )
        regen_coldkey_parser.add_argument(
            "--use_password",
            dest="use_password",
            action="store_true",
            help="""Set true to protect the generated nimble key with a password.""",
            default=True,
        )
        regen_coldkey_parser.add_argument(
            "--no_password",
            dest="use_password",
            action="store_false",
            help="""Set off protects the generated nimble key with a password.""",
        )
        regen_coldkey_parser.add_argument(
            "--overwrite_coldkey",
            default=False,
            action="store_false",
            help="""Overwrite the old coldkey with the newly generated coldkey""",
        )
        nimble.wallet.add_args(regen_coldkey_parser)
        nimble.nbnetwork.add_args(regen_coldkey_parser)


class RegenColdkeypubCommand:
    """
    Executes the 'regen_coldkeypub' command to regenerate the public part of a coldkey (coldkeypub) for a wallet on the nimble network.
    This command is used when a user needs to recreate their coldkeypub from an existing public key or SS58 address.

    Usage:
    The command requires either a public key in hexadecimal format or an SS58 address to regenerate the coldkeypub.
    It optionally allows overwriting an existing coldkeypub file.

    Optional arguments:
    - --public_key_hex (str): The public key in hex format.
    - --ss58_address (str): The SS58 address of the coldkey.
    - --overwrite_coldkeypub (bool): Overwrites the existing coldkeypub file with the new one.

    Example usage:
    >>> nbcli wallet regen_coldkeypub --ss58_address 5DkQ4...

    Note:
    This command is particularly useful for users who need to regenerate their coldkeypub, perhaps due to file corruption or loss.
    It is a recovery-focused utility that ensures continued access to wallet functionalities.
    """

    def run(cli):
        r"""Creates a new coldkeypub under this wallet."""
        wallet = nimble.wallet(config=cli.config)
        wallet.regenerate_coldkeypub(
            ss58_address=cli.config.get("ss58_address"),
            public_key=cli.config.get("public_key_hex"),
            overwrite=cli.config.overwrite_coldkeypub,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)
        if config.ss58_address == None and config.public_key_hex == None:
            prompt_answer = Prompt.ask(
                "Enter the ss58_address or the public key in hex"
            )
            if prompt_answer.startswith("0x"):
                config.public_key_hex = prompt_answer
            else:
                config.ss58_address = prompt_answer
        if not nimble.utils.is_valid_nimble_address_or_public_key(
            address=config.ss58_address
            if config.ss58_address
            else config.public_key_hex
        ):
            sys.exit(1)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        regen_coldkeypub_parser = parser.add_parser(
            "regen_coldkeypub",
            help="""Regenerates a coldkeypub from the public part of the coldkey.""",
        )
        regen_coldkeypub_parser.add_argument(
            "--public_key",
            "--pubkey",
            dest="public_key_hex",
            required=False,
            default=None,
            type=str,
            help="The public key (in hex) of the coldkey to regen e.g. 0x1234 ...",
        )
        regen_coldkeypub_parser.add_argument(
            "--ss58_address",
            "--addr",
            "--ss58",
            dest="ss58_address",
            required=False,
            default=None,
            type=str,
            help="The ss58 address of the coldkey to regen e.g. 5ABCD ...",
        )
        regen_coldkeypub_parser.add_argument(
            "--overwrite_coldkeypub",
            default=False,
            action="store_true",
            help="""Overwrite the old coldkeypub file with the newly generated coldkeypub""",
        )
        nimble.wallet.add_args(regen_coldkeypub_parser)
        nimble.nbnetwork.add_args(regen_coldkeypub_parser)


class RegenHotkeyCommand:
    """
    Executes the 'regen_hotkey' command to regenerate a hotkey for a wallet on the nimble network.
    Similar to regenerating a coldkey, this command creates a new hotkey from a mnemonic, seed, or JSON file.

    Usage:
    Users can provide a mnemonic, seed string, or a JSON file to regenerate the hotkey.
    The command supports optional password protection and can overwrite an existing hotkey.

    Optional arguments:
    - --mnemonic (str): A mnemonic phrase used to regenerate the key.
    - --seed (str): A seed hex string used for key regeneration.
    - --json (str): Path to a JSON file containing an encrypted key backup.
    - --json_password (str): Password to decrypt the JSON file.
    - --use_password (bool): Enables password protection for the generated key.
    - --overwrite_hotkey (bool): Overwrites the existing hotkey with the new one.

    Example usage:
    >>> nbcli wallet regen_hotkey
    >>> nbcli wallet regen_hotkey --seed 0x1234...

    Note:
    This command is essential for users who need to regenerate their hotkey, possibly for security upgrades or key recovery.
    It should be used cautiously to avoid accidental overwrites of existing keys.
    """

    def run(cli):
        r"""Creates a new coldkey under this wallet."""
        wallet = nimble.wallet(config=cli.config)

        json_str: Optional[str] = None
        json_password: Optional[str] = None
        if cli.config.get("json"):
            file_name: str = cli.config.get("json")
            if not os.path.exists(file_name) or not os.path.isfile(file_name):
                raise ValueError("File {} does not exist".format(file_name))
            with open(cli.config.get("json"), "r") as f:
                json_str = f.read()

            # Password can be "", assume if None
            json_password = cli.config.get("json_password", "")

        wallet.regenerate_hotkey(
            mnemonic=cli.config.mnemonic,
            seed=cli.config.seed,
            json=(json_str, json_password),
            use_password=cli.config.use_password,
            overwrite=cli.config.overwrite_hotkey,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

        if not config.is_set("wallet.hotkey") and not config.no_predict:
            hotkey = Prompt.ask(
                "Enter hotkey name", default=defaults.wallet.hotkey
            )
            config.wallet.hotkey = str(hotkey)
        if (
            config.mnemonic == None
            and config.get("seed", d=None) == None
            and config.get("json", d=None) == None
        ):
            prompt_answer = Prompt.ask(
                "Enter mnemonic, seed, or json file location"
            )
            if prompt_answer.startswith("0x"):
                config.seed = prompt_answer
            elif len(prompt_answer.split(" ")) > 1:
                config.mnemonic = prompt_answer
            else:
                config.json = prompt_answer

        if (
            config.get("json", d=None)
            and config.get("json_password", d=None) == None
        ):
            config.json_password = Prompt.ask(
                "Enter json backup password", password=True
            )

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        regen_hotkey_parser = parser.add_parser(
            "regen_hotkey",
            help="""Regenerates a hotkey from a passed mnemonic""",
        )
        regen_hotkey_parser.add_argument(
            "--mnemonic",
            required=False,
            nargs="+",
            help="Mnemonic used to regen your key i.e. horse cart dog ...",
        )
        regen_hotkey_parser.add_argument(
            "--seed",
            required=False,
            default=None,
            help="Seed hex string used to regen your key i.e. 0x1234...",
        )
        regen_hotkey_parser.add_argument(
            "--json",
            required=False,
            default=None,
            help="""Path to a json file containing the encrypted key backup. (e.g. from PolkadotJS)""",
        )
        regen_hotkey_parser.add_argument(
            "--json_password",
            required=False,
            default=None,
            help="""Password to decrypt the json file.""",
        )
        regen_hotkey_parser.add_argument(
            "--use_password",
            dest="use_password",
            action="store_true",
            help="""Set true to protect the generated nimble key with a password.""",
            default=False,
        )
        regen_hotkey_parser.add_argument(
            "--no_password",
            dest="use_password",
            action="store_false",
            help="""Set off protects the generated nimble key with a password.""",
        )
        regen_hotkey_parser.add_argument(
            "--overwrite_hotkey",
            dest="overwrite_hotkey",
            action="store_true",
            default=False,
            help="""Overwrite the old hotkey with the newly generated hotkey""",
        )
        nimble.wallet.add_args(regen_hotkey_parser)
        nimble.nbnetwork.add_args(regen_hotkey_parser)


class NewHotkeyCommand:
    """
    Executes the 'new_hotkey' command to create a new hotkey under a wallet on the nimble network.
    This command is used to generate a new hotkey for managing a particle or participating in the network.

    Usage:
    The command creates a new hotkey with an optional word count for the mnemonic and supports password protection.
    It also allows overwriting an existing hotkey.

    Optional arguments:
    - --n_words (int): The number of words in the mnemonic phrase.
    - --use_password (bool): Enables password protection for the generated key.
    - --overwrite_hotkey (bool): Overwrites the existing hotkey with the new one.

    Example usage:
    >>> nbcli wallet new_hotkey --n_words 24

    Note:
    This command is useful for users who wish to create additional hotkeys for different purposes,
    such as running multiple miners or separating operational roles within the network.
    """

    def run(cli):
        """Creates a new hotke under this wallet."""
        wallet = nimble.wallet(config=cli.config)
        wallet.create_new_hotkey(
            n_words=cli.config.n_words,
            use_password=cli.config.use_password,
            overwrite=cli.config.overwrite_hotkey,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

        if not config.is_set("wallet.hotkey") and not config.no_predict:
            hotkey = Prompt.ask(
                "Enter hotkey name", default=defaults.wallet.hotkey
            )
            config.wallet.hotkey = str(hotkey)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        new_hotkey_parser = parser.add_parser(
            "new_hotkey",
            help="""Creates a new hotkey (for running a miner) under the specified path.""",
        )
        new_hotkey_parser.add_argument(
            "--n_words",
            type=int,
            choices=[12, 15, 18, 21, 24],
            default=12,
            help="""The number of words representing the mnemonic. i.e. horse cart dog ... x 24""",
        )
        new_hotkey_parser.add_argument(
            "--use_password",
            dest="use_password",
            action="store_true",
            help="""Set true to protect the generated nimble key with a password.""",
            default=False,
        )
        new_hotkey_parser.add_argument(
            "--no_password",
            dest="use_password",
            action="store_false",
            help="""Set off protects the generated nimble key with a password.""",
        )
        new_hotkey_parser.add_argument(
            "--overwrite_hotkey",
            action="store_false",
            default=False,
            help="""Overwrite the old hotkey with the newly generated hotkey""",
        )
        nimble.wallet.add_args(new_hotkey_parser)
        nimble.nbnetwork.add_args(new_hotkey_parser)


class NewColdkeyCommand:
    """
    Executes the 'new_coldkey' command to create a new coldkey under a wallet on the nimble network.
    This command generates a coldkey, which is essential for holding balances and performing high-value transactions.

    Usage:
    The command creates a new coldkey with an optional word count for the mnemonic and supports password protection.
    It also allows overwriting an existing coldkey.

    Optional arguments:
    - --n_words (int): The number of words in the mnemonic phrase.
    - --use_password (bool): Enables password protection for the generated key.
    - --overwrite_coldkey (bool): Overwrites the existing coldkey with the new one.

    Example usage:
    >>> nbcli wallet new_coldkey --n_words 15

    Note:
    This command is crucial for users who need to create a new coldkey for enhanced security or as part of setting up a new wallet.
    It's a foundational step in establishing a secure presence on the nimble network.
    """

    def run(cli):
        r"""Creates a new coldkey under this wallet."""
        wallet = nimble.wallet(config=cli.config)
        wallet.create_new_coldkey(
            n_words=cli.config.n_words,
            use_password=cli.config.use_password,
            overwrite=cli.config.overwrite_coldkey,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        new_coldkey_parser = parser.add_parser(
            "new_coldkey",
            help="""Creates a new coldkey (for containing balance) under the specified path. """,
        )
        new_coldkey_parser.add_argument(
            "--n_words",
            type=int,
            choices=[12, 15, 18, 21, 24],
            default=12,
            help="""The number of words representing the mnemonic. i.e. horse cart dog ... x 24""",
        )
        new_coldkey_parser.add_argument(
            "--use_password",
            dest="use_password",
            action="store_true",
            help="""Set true to protect the generated nimble key with a password.""",
            default=True,
        )
        new_coldkey_parser.add_argument(
            "--no_password",
            dest="use_password",
            action="store_false",
            help="""Set off protects the generated nimble key with a password.""",
        )
        new_coldkey_parser.add_argument(
            "--overwrite_coldkey",
            action="store_false",
            default=False,
            help="""Overwrite the old coldkey with the newly generated coldkey""",
        )
        nimble.wallet.add_args(new_coldkey_parser)
        nimble.nbnetwork.add_args(new_coldkey_parser)


class WalletCreateCommand:
    """
    Executes the 'create' command to generate both a new coldkey and hotkey under a specified wallet on the nimble network.
    This command is a comprehensive utility for creating a complete wallet setup with both cold and hotkeys.

    Usage:
    The command facilitates the creation of a new coldkey and hotkey with an optional word count for the mnemonics.
    It supports password protection for the coldkey and allows overwriting of existing keys.

    Optional arguments:
    - --n_words (int): The number of words in the mnemonic phrase for both keys.
    - --use_password (bool): Enables password protection for the coldkey.
    - --overwrite_coldkey (bool): Overwrites the existing coldkey with the new one.
    - --overwrite_hotkey (bool): Overwrites the existing hotkey with the new one.

    Example usage:
    >>> nbcli wallet create --n_words 21

    Note:
    This command is ideal for new users setting up their wallet for the first time or for those who wish to completely renew their wallet keys.
    It ensures a fresh start with new keys for secure and effective participation in the network.
    """

    def run(cli):
        r"""Creates a new coldkey and hotkey under this wallet."""
        wallet = nimble.wallet(config=cli.config)
        wallet.create_new_coldkey(
            n_words=cli.config.n_words,
            use_password=cli.config.use_password,
            overwrite=cli.config.overwrite_coldkey,
        )
        wallet.create_new_hotkey(
            n_words=cli.config.n_words,
            use_password=False,
            overwrite=cli.config.overwrite_hotkey,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)
        if not config.is_set("wallet.hotkey") and not config.no_predict:
            hotkey = Prompt.ask(
                "Enter hotkey name", default=defaults.wallet.hotkey
            )
            config.wallet.hotkey = str(hotkey)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        new_coldkey_parser = parser.add_parser(
            "create",
            help="""Creates a new coldkey (for containing balance) under the specified path. """,
        )
        new_coldkey_parser.add_argument(
            "--n_words",
            type=int,
            choices=[12, 15, 18, 21, 24],
            default=12,
            help="""The number of words representing the mnemonic. i.e. horse cart dog ... x 24""",
        )
        new_coldkey_parser.add_argument(
            "--use_password",
            dest="use_password",
            action="store_true",
            help="""Set true to protect the generated nimble key with a password.""",
            default=True,
        )
        new_coldkey_parser.add_argument(
            "--no_password",
            dest="use_password",
            action="store_false",
            help="""Set off protects the generated nimble key with a password.""",
        )
        new_coldkey_parser.add_argument(
            "--overwrite_coldkey",
            action="store_false",
            default=False,
            help="""Overwrite the old coldkey with the newly generated coldkey""",
        )
        new_coldkey_parser.add_argument(
            "--overwrite_hotkey",
            action="store_false",
            default=False,
            help="""Overwrite the old hotkey with the newly generated hotkey""",
        )
        nimble.wallet.add_args(new_coldkey_parser)
        nimble.nbnetwork.add_args(new_coldkey_parser)


def _get_coldkey_wallets_for_path(path: str) -> List["nimble.wallet"]:
    """Get all coldkey wallet names from path."""
    try:
        wallet_names = next(os.walk(os.path.expanduser(path)))[1]
        return [nimble.wallet(path=path, name=name) for name in wallet_names]
    except StopIteration:
        # No wallet files found.
        wallets = []
    return wallets


class UpdateWalletCommand:
    """
    Executes the 'update' command to check and potentially update the security of the wallets in the nimble network.
    This command is used to enhance wallet security using modern encryption standards.

    Usage:
    The command checks if any of the wallets need an update in their security protocols.
    It supports updating all legacy wallets or a specific one based on the user's choice.

    Optional arguments:
    - --all (bool): When set, updates all legacy wallets.
    - --no_predict (bool): Disables user inference during the update process.

    Example usage:
    >>> nbcli wallet update --all

    Note:
    This command is important for maintaining the highest security standards for users' wallets.
    It is recommended to run this command periodically to ensure wallets are up-to-date with the latest security practices.
    """

    @staticmethod
    def run(cli):
        """Check if any of the wallets needs an update."""
        config = cli.config.copy()
        if config.get("all", d=False) == True:
            wallets = _get_coldkey_wallets_for_path(config.wallet.path)
        else:
            wallets = [nimble.wallet(config=config)]

        for wallet in wallets:
            print("\n===== ", wallet, " =====")
            wallet.coldkey_file.check_and_update_encryption()

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        update_wallet_parser = parser.add_parser(
            "update",
            help="""Updates the wallet security using NaCL instead of anvisible vault.""",
        )
        update_wallet_parser.add_argument("--all", action="store_true")
        update_wallet_parser.add_argument(
            "--no_predict",
            dest="no_predict",
            action="store_true",
            help="""Set true to avoid inference the user.""",
            default=False,
        )
        nimble.wallet.add_args(update_wallet_parser)
        nimble.nbnetwork.add_args(update_wallet_parser)

    @staticmethod
    def check_config(config: "nimble.Config"):
        if config.get("all", d=False) == False:
            if not config.no_predict:
                if Confirm.ask("Do you want to update all legacy wallets?"):
                    config["all"] = True

        # Ask the user to specify the wallet if the wallet name is not clear.
        if (
            config.get("all", d=False) == False
            and config.wallet.get("name") == nimble.defaults.wallet.name
            and not config.no_predict
        ):
            wallet_name = Prompt.ask(
                "Enter wallet name", default=nimble.defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)


def _get_coldkey_ss58_addresses_for_path(path: str) -> List[str]:
    """Get all coldkey ss58 addresses from path."""

    def list_coldkeypub_files(dir_path):
        abspath = os.path.abspath(os.path.expanduser(dir_path))
        coldkey_files = []

        for file in os.listdir(abspath):
            coldkey_path = os.path.join(abspath, file, "coldkeypub.txt")
            if os.path.exists(coldkey_path):
                coldkey_files.append(coldkey_path)
            else:
                nimble.logging.warning(
                    f"{coldkey_path} does not exist. Excluding..."
                )
        return coldkey_files

    return [
        nimble.keyfile(file).keypair.ss58_address
        for file in list_coldkeypub_files(path)
    ]


class WalletBalanceCommand:
    """
    Executes the 'balance' command to check the balance of the wallet on the nimble network.
    This command provides a detailed view of the wallet's coldkey balances, including free and staked balances.

    Usage:
    The command lists the balances of all wallets in the user's configuration directory, showing the wallet name,
    coldkey address, and the respective free and staked balances.

    Optional arguments:
    None. The command uses the wallet and nbnetwork configurations to fetch balance data.

    Example usage:
    >>> nbcli wallet balance

    Note:
    This command is essential for users to monitor their financial status on the nimble network.
    It helps in keeping track of assets and ensuring the wallet's financial health.
    """

    @staticmethod
    def run(cli):
        """Check the balance of the wallet."""
        wallet_names = os.listdir(os.path.expanduser(cli.config.wallet.path))
        coldkeys = _get_coldkey_ss58_addresses_for_path(cli.config.wallet.path)
        nbnetwork = nimble.nbnetwork(config=cli.config, log_verbose=False)

        free_balances = [
            nbnetwork.get_balance(coldkeys[i]) for i in range(len(coldkeys))
        ]
        staked_balances = [
            nbnetwork.get_total_stake_for_coldkey(coldkeys[i])
            for i in range(len(coldkeys))
        ]
        total_free_balance = sum(free_balances)
        total_staked_balance = sum(staked_balances)

        balances = {
            name: (coldkey, free, staked)
            for name, coldkey, free, staked in sorted(
                zip(wallet_names, coldkeys, free_balances, staked_balances)
            )
        }

        table = Table(show_footer=False)
        table.title = "[white]Wallet Coldkey Balances"
        table.add_column(
            "[white]Wallet Name",
            header_style="overline white",
            footer_style="overline white",
            style="rgb(50,163,219)",
            no_wrap=True,
        )

        table.add_column(
            "[white]Coldkey Address",
            header_style="overline white",
            footer_style="overline white",
            style="rgb(50,163,219)",
            no_wrap=True,
        )

        for typestr in ["Free", "Staked", "Total"]:
            table.add_column(
                f"[white]{typestr} Balance",
                header_style="overline white",
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )

        for name, (coldkey, free, staked) in balances.items():
            table.add_row(
                name,
                coldkey,
                str(free),
                str(staked),
                str(free + staked),
            )
        table.add_row()
        table.add_row(
            "Total Balance Across All Coldkeys",
            "",
            str(total_free_balance),
            str(total_staked_balance),
            str(total_free_balance + total_staked_balance),
        )
        table.show_footer = True

        table.box = None
        table.pad_edge = False
        table.width = None
        nimble.__console__.print(table)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        balance_parser = parser.add_parser(
            "balance", help="""Checks the balance of the wallet."""
        )
        nimble.wallet.add_args(balance_parser)
        nimble.nbnetwork.add_args(balance_parser)

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.path") and not config.no_predict:
            path = Prompt.ask(
                "Enter wallets path", default=defaults.wallet.path
            )
            config.wallet.path = str(path)

        if not config.is_set("nbnetwork.network") and not config.no_predict:
            network = Prompt.ask(
                "Enter network",
                default=defaults.nbnetwork.network,
                choices=nimble.__networks__,
            )
            config.nbnetwork.network = str(network)
            (
                _,
                config.nbnetwork.chain_endpoint,
            ) = nimble.nbnetwork.determine_chain_endpoint_and_network(
                str(network)
            )



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
import shtab
import argparse
import nimble
from typing import List, Optional
from .commands import *

# Create a console instance for CLI display.
console = nimble.__console__

ALIAS_TO_COMMAND = {
    "subnets": "subnets",
    "s": "subnets",
    "wallet": "wallet",
    "stake": "stake",
    "wallets": "wallet",
    "w": "wallet",
    "st": "stake",
}
COMMANDS = {
    "wallet": {
        "name": "wallet",
        "aliases": ["w", "wallets"],
        "help": "Commands for managing and viewing wallets.",
        "commands": {
            "list": ListCommand,
            "overview": OverviewCommand,
            "transfer": TransferCommand,
            "create": WalletCreateCommand,
            "new_hotkey": NewHotkeyCommand,
            "new_coldkey": NewColdkeyCommand
        },
    },
    "subnets": {
        "name": "subnets",
        "aliases": ["s", "subnet"],
        "help": "Commands for managing and viewing subnetworks.",
        "commands": {
            "register": RegisterCommand,
        },
    },
    "stake": {
        "name": "stake",
        "aliases": ["st", "stakes"],
        "help": "Commands for staking and removing stake from hotkey accounts.",
        "commands": {
            "show": StakeShow,
            "add": StakeCommand,
            "remove": UnStakeCommand
        },
    }
}


class CLIErrorParser(argparse.ArgumentParser):
    """
    Custom ArgumentParser for better error messages.
    """

    def error(self, message):
        """
        This method is called when an error occurs. It prints a custom error message.
        """
        sys.stderr.write(f"Error: {message}\n")
        self.print_help()
        sys.exit(2)


class cli:
    """
    Implementation of the Command Line Interface (CLI) class for the nimble protocol.
    This class handles operations like key management (hotkey and coldkey) and token transfer.
    """

    def __init__(
        self,
        config: Optional["nimble.config"] = None,
        args: Optional[List[str]] = None,
    ):
        """
        Initializes a nimble.CLI object.

        Args:
            config (nimble.config, optional): The configuration settings for the CLI.
            args (List[str], optional): List of command line arguments.
        """
        # Turns on console for cli.
        nimble.turn_console_on()

        # If no config is provided, create a new one from args.
        if config == None:
            config = cli.create_config(args)

        self.config = config
        if self.config.command in ALIAS_TO_COMMAND:
            self.config.command = ALIAS_TO_COMMAND[self.config.command]
        else:
            console.print(
                f":cross_mark:[red]Unknown command: {self.config.command}[/red]"
            )
            sys.exit()

        # Check if the config is valid.
        cli.check_config(self.config)

        # If no_version_checking is not set or set as False in the config, version checking is done.
        if not self.config.get("no_version_checking", d=True):
            try:
                nimble.utils.version_checking()
            except:
                # If version checking fails, inform user with an exception.
                raise RuntimeError(
                    "To avoid internet-based version checking, pass --no_version_checking while running the CLI."
                )

    @staticmethod
    def __create_parser__() -> "argparse.ArgumentParser":
        """
        Creates the argument parser for the nimble CLI.

        Returns:
            argparse.ArgumentParser: An argument parser object for nimble CLI.
        """
        # Define the basic argument parser.
        parser = CLIErrorParser(
            description=f"nimble cli v{nimble.__version__}",
            usage="nbcli <command> <command args>",
            add_help=True,
        )
        # Add shtab completion
        parser.add_argument(
            "--print-completion",
            choices=shtab.SUPPORTED_SHELLS,
            help="Print shell tab completion script",
        )
        # Add arguments for each sub-command.
        cmd_parsers = parser.add_subparsers(dest="command")
        # Add argument parsers for all available commands.
        for command in COMMANDS.values():
            if isinstance(command, dict):
                subcmd_parser = cmd_parsers.add_parser(
                    name=command["name"],
                    aliases=command["aliases"],
                    help=command["help"],
                )
                subparser = subcmd_parser.add_subparsers(
                    help=command["help"], dest="subcommand", required=True
                )

                for subcommand in command["commands"].values():
                    subcommand.add_args(subparser)
            else:
                command.add_args(cmd_parsers)

        return parser

    @staticmethod
    def create_config(args: List[str]) -> "nimble.config":
        """
        From the argument parser, add config to nimble.executor and local config

        Args:
            args (List[str]): List of command line arguments.

        Returns:
            nimble.config: The configuration object for nimble CLI.
        """
        parser = cli.__create_parser__()

        # If no arguments are passed, print help text and exit the program.
        if len(args) == 0:
            parser.print_help()
            sys.exit()

        return nimble.config(parser, args=args)

    @staticmethod
    def check_config(config: "nimble.config"):
        """
        Checks if the essential configuration exists under different command

        Args:
            config (nimble.config): The configuration settings for the CLI.
        """
        # Check if command exists, if so, run the corresponding check_config.
        # If command doesn't exist, inform user and exit the program.
        if config.command in COMMANDS:
            command = config.command
            command_data = COMMANDS[command]

            if isinstance(command_data, dict):
                if config["subcommand"] != None:
                    command_data["commands"][config["subcommand"]].check_config(
                        config
                    )
                else:
                    console.print(
                        f":cross_mark:[red]Missing subcommand for: {config.command}[/red]"
                    )
                    sys.exit(1)
            else:
                command_data.check_config(config)
        else:
            console.print(
                f":cross_mark:[red]Unknown command: {config.command}[/red]"
            )
            sys.exit(1)

    def run(self):
        """
        Executes the command from the configuration.
        """
        # Check for print-completion argument
        if self.config.print_completion:
            shell = self.config.print_completion
            print(shtab.complete(parser, shell))
            return

        # Check if command exists, if so, run the corresponding method.
        # If command doesn't exist, inform user and exit the program.
        command = self.config.command
        if command in COMMANDS:
            command_data = COMMANDS[command]

            if isinstance(command_data, dict):
                command_data["commands"][self.config["subcommand"]].run(self)
            else:
                command_data.run(self)
        else:
            console.print(
                f":cross_mark:[red]Unknown command: {self.config.command}[/red]"
            )
            sys.exit()



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
import torch
import nimble

import json
from enum import Enum
from dataclasses import dataclass, asdict
from scalecodec.types import GenericCall
from typing import List, Tuple, Dict, Optional, Any, TypedDict, Union
from scalecodec.base import RuntimeConfiguration, ScaleBytes
from scalecodec.type_registry import load_type_registry_preset
from scalecodec.utils.ss58 import ss58_encode

from .utils import networking as net, U16_MAX, U16_NORMALIZED_FLOAT
from .utils.balance import Balance

custom_rpc_type_registry = {
    "types": {
        "CosmosInfo": {
            "type": "struct",
            "type_mapping": [
                ["netuid", "Compact<u16>"],
                ["rho", "Compact<u16>"],
                ["kappa", "Compact<u16>"],
                ["difficulty", "Compact<u64>"],
                ["immunity_period", "Compact<u16>"],
                ["max_allowed_validators", "Compact<u16>"],
                ["min_allowed_weights", "Compact<u16>"],
                ["max_weights_limit", "Compact<u16>"],
                ["scaling_law_power", "Compact<u16>"],
                ["cosmos_n", "Compact<u16>"],
                ["max_allowed_uids", "Compact<u16>"],
                ["blocks_since_last_step", "Compact<u64>"],
                ["tempo", "Compact<u16>"],
                ["network_modality", "Compact<u16>"],
                ["network_connect", "Vec<[u16; 2]>"],
                ["emission_values", "Compact<u64>"],
                ["burn", "Compact<u64>"],
                ["owner", "AccountId"],
            ],
        },
        "DelegateInfo": {
            "type": "struct",
            "type_mapping": [
                ["delegate_ss58", "AccountId"],
                ["take", "Compact<u16>"],
                ["nominators", "Vec<(AccountId, Compact<u64>)>"],
                ["owner_ss58", "AccountId"],
                ["registrations", "Vec<Compact<u16>>"],
                ["validator_permits", "Vec<Compact<u16>>"],
                ["return_per_1000", "Compact<u64>"],
                ["total_daily_return", "Compact<u64>"],
            ],
        },
        "ParticleInfo": {
            "type": "struct",
            "type_mapping": [
                ["hotkey", "AccountId"],
                ["coldkey", "AccountId"],
                ["uid", "Compact<u16>"],
                ["netuid", "Compact<u16>"],
                ["active", "bool"],
                ["fermion_info", "fermion_info"],
                ["proton_info", "ProtonInfo"],
                ["stake", "Vec<(AccountId, Compact<u64>)>"],
                ["rank", "Compact<u16>"],
                ["emission", "Compact<u64>"],
                ["incentive", "Compact<u16>"],
                ["consensus", "Compact<u16>"],
                ["trust", "Compact<u16>"],
                ["validator_trust", "Compact<u16>"],
                ["dividends", "Compact<u16>"],
                ["last_update", "Compact<u64>"],
                ["validator_permit", "bool"],
                ["weights", "Vec<(Compact<u16>, Compact<u16>)>"],
                ["bonds", "Vec<(Compact<u16>, Compact<u16>)>"],
                ["pruning_score", "Compact<u16>"],
            ],
        },
        "ParticleInfoLite": {
            "type": "struct",
            "type_mapping": [
                ["hotkey", "AccountId"],
                ["coldkey", "AccountId"],
                ["uid", "Compact<u16>"],
                ["netuid", "Compact<u16>"],
                ["active", "bool"],
                ["fermion_info", "fermion_info"],
                ["proton_info", "ProtonInfo"],
                ["stake", "Vec<(AccountId, Compact<u64>)>"],
                ["rank", "Compact<u16>"],
                ["emission", "Compact<u64>"],
                ["incentive", "Compact<u16>"],
                ["consensus", "Compact<u16>"],
                ["trust", "Compact<u16>"],
                ["validator_trust", "Compact<u16>"],
                ["dividends", "Compact<u16>"],
                ["last_update", "Compact<u64>"],
                ["validator_permit", "bool"],
                ["pruning_score", "Compact<u16>"],
            ],
        },
        "fermion_info": {
            "type": "struct",
            "type_mapping": [
                ["block", "u64"],
                ["version", "u32"],
                ["ip", "u128"],
                ["port", "u16"],
                ["ip_type", "u8"],
                ["protocol", "u8"],
                ["placeholder1", "u8"],
                ["placeholder2", "u8"],
            ],
        },
        "ProtonInfo": {
            "type": "struct",
            "type_mapping": [
                ["block", "u64"],
                ["version", "u32"],
                ["ip", "u128"],
                ["port", "u16"],
                ["ip_type", "u8"],
            ],
        },
        "IPInfo": {
            "type": "struct",
            "type_mapping": [
                ["ip", "Compact<u128>"],
                ["ip_type_and_protocol", "Compact<u8>"],
            ],
        },
        "StakeInfo": {
            "type": "struct",
            "type_mapping": [
                ["hotkey", "AccountId"],
                ["coldkey", "AccountId"],
                ["stake", "Compact<u64>"],
            ],
        },
        "CosmosHyperparameters": {
            "type": "struct",
            "type_mapping": [
                ["rho", "Compact<u16>"],
                ["kappa", "Compact<u16>"],
                ["immunity_period", "Compact<u16>"],
                ["min_allowed_weights", "Compact<u16>"],
                ["max_weights_limit", "Compact<u16>"],
                ["tempo", "Compact<u16>"],
                ["min_difficulty", "Compact<u64>"],
                ["max_difficulty", "Compact<u64>"],
                ["weights_version", "Compact<u64>"],
                ["weights_rate_limit", "Compact<u64>"],
                ["adjustment_interval", "Compact<u16>"],
                ["activity_cutoff", "Compact<u16>"],
                ["registration_allowed", "bool"],
                ["target_regs_per_interval", "Compact<u16>"],
                ["min_burn", "Compact<u64>"],
                ["max_burn", "Compact<u64>"],
                ["bonds_moving_avg", "Compact<u64>"],
                ["max_regs_per_block", "Compact<u16>"],
                ["serving_rate_limit", "Compact<u64>"],
                ["max_validators", "Compact<u16>"],
            ],
        },
    }
}


@dataclass
class FermionInfo:
    version: int
    ip: str
    port: int
    ip_type: int
    hotkey: str
    coldkey: str
    protocol: int = 4
    placeholder1: int = 0
    placeholder2: int = 0

    @property
    def is_serving(self) -> bool:
        """True if the endpoint is serving."""
        if self.ip == "0.0.0.0":
            return False
        else:
            return True

    def ip_str(self) -> str:
        """Return the whole ip as string"""
        return net.ip__str__(self.ip_type, self.ip, self.port)

    def __eq__(self, other: "FermionInfo"):
        if other == None:
            return False
        if (
            self.version == other.version
            and self.ip == other.ip
            and self.port == other.port
            and self.ip_type == other.ip_type
            and self.coldkey == other.coldkey
            and self.hotkey == other.hotkey
        ):
            return True
        else:
            return False

    def __str__(self):
        return "FermionInfo( {}, {}, {}, {} )".format(
            str(self.ip_str()),
            str(self.hotkey),
            str(self.coldkey),
            self.version,
        )

    def __repr__(self):
        return self.__str__()

    def to_string(self) -> str:
        """Converts the FermionInfo object to a string representation using JSON."""
        try:
            return json.dumps(asdict(self))
        except (TypeError, ValueError) as e:
            nimble.logging.error(f"Error converting FermionInfo to string: {e}")
            return FermionInfo(0, "", 0, 0, "", "").to_string()

    @classmethod
    def from_string(cls, s: str) -> "FermionInfo":
        """Creates an FermionInfo object from its string representation using JSON."""
        try:
            data = json.loads(s)
            return cls(**data)
        except json.JSONDecodeError as e:
            nimble.logging.error(f"Error decoding JSON: {e}")
        except TypeError as e:
            nimble.logging.error(f"Type error: {e}")
        except ValueError as e:
            nimble.logging.error(f"Value error: {e}")
        return FermionInfo(0, "", 0, 0, "", "")

    @classmethod
    def from_particle_info(cls, particle_info: dict) -> "FermionInfo":
        """Converts a dictionary to an fermion_info object."""
        return cls(
            version=particle_info["fermion_info"]["version"],
            ip=net.int_to_ip(int(particle_info["fermion_info"]["ip"])),
            port=particle_info["fermion_info"]["port"],
            ip_type=particle_info["fermion_info"]["ip_type"],
            hotkey=particle_info["hotkey"],
            coldkey=particle_info["coldkey"],
        )

    def to_parameter_dict(self) -> "torch.nn.ParameterDict":
        r"""Returns a torch tensor of the cosmos info."""
        return torch.nn.ParameterDict(self.__dict__)

    @classmethod
    def from_parameter_dict(
        cls, parameter_dict: "torch.nn.ParameterDict"
    ) -> "fermion_info":
        r"""Returns an fermion_info object from a torch parameter_dict."""
        return cls(**dict(parameter_dict))


class ChainDataType(Enum):
    ParticleInfo = 1
    CosmosInfo = 2
    DelegateInfo = 3
    ParticleInfoLite = 4
    DelegatedInfo = 5
    StakeInfo = 6
    IPInfo = 7
    CosmosHyperparameters = 8


# Constants
VIMPERNIM = 1e9
U16_MAX = 65535
U64_MAX = 18446744073709551615


def from_scale_encoding(
    input: Union[List[int], bytes, ScaleBytes],
    type_name: ChainDataType,
    is_vec: bool = False,
    is_option: bool = False,
) -> Optional[Dict]:
    type_string = type_name.name
    if type_name == ChainDataType.DelegatedInfo:
        # DelegatedInfo is a tuple of (DelegateInfo, Compact<u64>)
        type_string = f"({ChainDataType.DelegateInfo.name}, Compact<u64>)"
    if is_option:
        type_string = f"Option<{type_string}>"
    if is_vec:
        type_string = f"Vec<{type_string}>"

    return from_scale_encoding_using_type_string(input, type_string)


def from_scale_encoding_using_type_string(
    input: Union[List[int], bytes, ScaleBytes], type_string: str
) -> Optional[Dict]:
    if isinstance(input, ScaleBytes):
        as_scale_bytes = input
    else:
        if isinstance(input, list) and all([isinstance(i, int) for i in input]):
            vec_u8 = input
            as_bytes = bytes(vec_u8)
        elif isinstance(input, bytes):
            as_bytes = input
        else:
            raise TypeError("input must be a List[int], bytes, or ScaleBytes")

        as_scale_bytes = ScaleBytes(as_bytes)

    rpc_runtime_config = RuntimeConfiguration()
    rpc_runtime_config.update_type_registry(load_type_registry_preset("legacy"))
    rpc_runtime_config.update_type_registry(custom_rpc_type_registry)

    obj = rpc_runtime_config.create_scale_object(
        type_string, data=as_scale_bytes
    )

    return obj.decode()


# Dataclasses for chain data.
@dataclass
class ParticleInfo:
    r"""
    Dataclass for particle metadata.
    """
    hotkey: str
    coldkey: str
    uid: int
    netuid: int
    active: int
    stake: Balance
    # mapping of coldkey to amount staked to this Particle
    stake_dict: Dict[str, Balance]
    total_stake: Balance
    rank: float
    emission: float
    incentive: float
    consensus: float
    trust: float
    validator_trust: float
    dividends: float
    last_update: int
    validator_permit: bool
    weights: List[List[int]]
    bonds: List[List[int]]
    proton_info: "ProtonInfo"
    fermion_info: "FermionInfo"
    pruning_score: int
    is_null: bool = False

    @classmethod
    def fix_decoded_values(cls, particle_info_decoded: Any) -> "ParticleInfo":
        r"""Fixes the values of the ParticleInfo object."""
        particle_info_decoded["hotkey"] = ss58_encode(
            particle_info_decoded["hotkey"], nimble.__ss58_format__
        )
        particle_info_decoded["coldkey"] = ss58_encode(
            particle_info_decoded["coldkey"], nimble.__ss58_format__
        )
        stake_dict = {
            ss58_encode(coldkey, nimble.__ss58_format__): Balance.from_vim(
                int(stake)
            )
            for coldkey, stake in particle_info_decoded["stake"]
        }
        particle_info_decoded["stake_dict"] = stake_dict
        particle_info_decoded["stake"] = sum(stake_dict.values())
        particle_info_decoded["total_stake"] = particle_info_decoded["stake"]
        particle_info_decoded["weights"] = [
            [int(weight[0]), int(weight[1])]
            for weight in particle_info_decoded["weights"]
        ]
        particle_info_decoded["bonds"] = [
            [int(bond[0]), int(bond[1])]
            for bond in particle_info_decoded["bonds"]
        ]
        particle_info_decoded["rank"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["rank"]
        )
        particle_info_decoded["emission"] = (
            particle_info_decoded["emission"] / VIMPERNIM
        )
        particle_info_decoded["incentive"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["incentive"]
        )
        particle_info_decoded["consensus"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["consensus"]
        )
        particle_info_decoded["trust"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["trust"]
        )
        particle_info_decoded["validator_trust"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["validator_trust"]
        )
        particle_info_decoded["dividends"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["dividends"]
        )
        particle_info_decoded[
            "proton_info"
        ] = ProtonInfo.fix_decoded_values(
            particle_info_decoded["proton_info"]
        )
        particle_info_decoded["fermion_info"] = FermionInfo.from_particle_info(
            particle_info_decoded
        )

        return cls(**particle_info_decoded)

    @classmethod
    def from_vec_u8(cls, vec_u8: List[int]) -> "ParticleInfo":
        r"""Returns a ParticleInfo object from a vec_u8."""
        if len(vec_u8) == 0:
            return ParticleInfo._null_particle()

        decoded = from_scale_encoding(vec_u8, ChainDataType.ParticleInfo)
        if decoded is None:
            return ParticleInfo._null_particle()

        decoded = ParticleInfo.fix_decoded_values(decoded)

        return decoded

    @classmethod
    def list_from_vec_u8(cls, vec_u8: List[int]) -> List["ParticleInfo"]:
        r"""Returns a list of ParticleInfo objects from a vec_u8."""

        decoded_list = from_scale_encoding(
            vec_u8, ChainDataType.ParticleInfo, is_vec=True
        )
        if decoded_list is None:
            return []

        decoded_list = [
            ParticleInfo.fix_decoded_values(decoded) for decoded in decoded_list
        ]
        return decoded_list

    @staticmethod
    def _null_particle() -> "ParticleInfo":
        particle = ParticleInfo(
            uid=0,
            netuid=0,
            active=0,
            stake=Balance.from_vim(0),
            stake_dict={},
            total_stake=Balance.from_vim(0),
            rank=0,
            emission=0,
            incentive=0,
            consensus=0,
            trust=0,
            validator_trust=0,
            dividends=0,
            last_update=0,
            validator_permit=False,
            weights=[],
            bonds=[],
            proton_info=None,
            fermion_info=None,
            is_null=True,
            coldkey="000000000000000000000000000000000000000000000000",
            hotkey="000000000000000000000000000000000000000000000000",
            pruning_score=0,
        )
        return particle

    @classmethod
    def from_weights_bonds_and_particle_lite(
        cls,
        particle_lite: "ParticleInfoLite",
        weights_as_dict: Dict[int, List[Tuple[int, int]]],
        bonds_as_dict: Dict[int, List[Tuple[int, int]]],
    ) -> "ParticleInfo":
        n_dict = particle_lite.__dict__
        n_dict["weights"] = weights_as_dict.get(particle_lite.uid, [])
        n_dict["bonds"] = bonds_as_dict.get(particle_lite.uid, [])

        return cls(**n_dict)

    @staticmethod
    def _particle_dict_to_namespace(particle_dict) -> "ParticleInfo":
        # TODO: Legacy: remove?
        if (
            particle_dict["hotkey"]
            == "5C4hrfjw9DjXZTzV3MwzrrAr9P1MJhSrvWGWqi1eSuyUpnhM"
        ):
            return ParticleInfo._null_particle()
        else:
            particle = ParticleInfo(**particle_dict)
            particle.stake_dict = {
                hk: Balance.from_vim(stake)
                for hk, stake in particle.stake.items()
            }
            particle.stake = Balance.from_vim(particle.total_stake)
            particle.total_stake = particle.stake
            particle.rank = particle.rank / U16_MAX
            particle.trust = particle.trust / U16_MAX
            particle.consensus = particle.consensus / U16_MAX
            particle.validator_trust = particle.validator_trust / U16_MAX
            particle.incentive = particle.incentive / U16_MAX
            particle.dividends = particle.dividends / U16_MAX
            particle.emission = particle.emission / VIMPERNIM

            return particle


@dataclass
class ParticleInfoLite:
    r"""
    Dataclass for particle metadata, but without the weights and bonds.
    """
    hotkey: str
    coldkey: str
    uid: int
    netuid: int
    active: int
    stake: Balance
    # mapping of coldkey to amount staked to this Particle
    stake_dict: Dict[str, Balance]
    total_stake: Balance
    rank: float
    emission: float
    incentive: float
    consensus: float
    trust: float
    validator_trust: float
    dividends: float
    last_update: int
    validator_permit: bool
    # weights: List[List[int]]
    # bonds: List[List[int]] No weights or bonds in lite version
    proton_info: "ProtonInfo"
    fermion_info: "fermion_info"
    pruning_score: int
    is_null: bool = False

    @classmethod
    def fix_decoded_values(cls, particle_info_decoded: Any) -> "ParticleInfoLite":
        r"""Fixes the values of the ParticleInfoLite object."""
        particle_info_decoded["hotkey"] = ss58_encode(
            particle_info_decoded["hotkey"], nimble.__ss58_format__
        )
        particle_info_decoded["coldkey"] = ss58_encode(
            particle_info_decoded["coldkey"], nimble.__ss58_format__
        )
        stake_dict = {
            ss58_encode(coldkey, nimble.__ss58_format__): Balance.from_vim(
                int(stake)
            )
            for coldkey, stake in particle_info_decoded["stake"]
        }
        particle_info_decoded["stake_dict"] = stake_dict
        particle_info_decoded["stake"] = sum(stake_dict.values())
        particle_info_decoded["total_stake"] = particle_info_decoded["stake"]
        # Don't need weights and bonds in lite version
        # particle_info_decoded['weights'] = [[int(weight[0]), int(weight[1])] for weight in particle_info_decoded['weights']]
        # particle_info_decoded['bonds'] = [[int(bond[0]), int(bond[1])] for bond in particle_info_decoded['bonds']]
        particle_info_decoded["rank"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["rank"]
        )
        particle_info_decoded["emission"] = (
            particle_info_decoded["emission"] / VIMPERNIM
        )
        particle_info_decoded["incentive"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["incentive"]
        )
        particle_info_decoded["consensus"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["consensus"]
        )
        particle_info_decoded["trust"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["trust"]
        )
        particle_info_decoded["validator_trust"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["validator_trust"]
        )
        particle_info_decoded["dividends"] = U16_NORMALIZED_FLOAT(
            particle_info_decoded["dividends"]
        )
        particle_info_decoded[
            "proton_info"
        ] = ProtonInfo.fix_decoded_values(
            particle_info_decoded["proton_info"]
        )
        particle_info_decoded["fermion_info"] = FermionInfo.from_particle_info(
            particle_info_decoded
        )
        return cls(**particle_info_decoded)

    @classmethod
    def from_vec_u8(cls, vec_u8: List[int]) -> "ParticleInfoLite":
        r"""Returns a ParticleInfoLite object from a vec_u8."""
        if len(vec_u8) == 0:
            return ParticleInfoLite._null_particle()

        decoded = from_scale_encoding(vec_u8, ChainDataType.ParticleInfoLite)
        if decoded is None:
            return ParticleInfoLite._null_particle()

        decoded = ParticleInfoLite.fix_decoded_values(decoded)

        return decoded

    @classmethod
    def list_from_vec_u8(cls, vec_u8: List[int]) -> List["ParticleInfoLite"]:
        r"""Returns a list of ParticleInfoLite objects from a vec_u8."""

        decoded_list = from_scale_encoding(
            vec_u8, ChainDataType.ParticleInfoLite, is_vec=True
        )
        if decoded_list is None:
            return []

        decoded_list = [
            ParticleInfoLite.fix_decoded_values(decoded)
            for decoded in decoded_list
        ]
        return decoded_list

    @staticmethod
    def _null_particle() -> "ParticleInfoLite":
        particle = ParticleInfoLite(
            uid=0,
            netuid=0,
            active=0,
            stake=Balance.from_vim(0),
            stake_dict={},
            total_stake=Balance.from_vim(0),
            rank=0,
            emission=0,
            incentive=0,
            consensus=0,
            trust=0,
            validator_trust=0,
            dividends=0,
            last_update=0,
            validator_permit=False,
            # weights = [], // No weights or bonds in lite version
            # bonds = [],
            proton_info=None,
            fermion_info=None,
            is_null=True,
            coldkey="000000000000000000000000000000000000000000000000",
            hotkey="000000000000000000000000000000000000000000000000",
            pruning_score=0,
        )
        return particle

    @staticmethod
    def _particle_dict_to_namespace(particle_dict) -> "ParticleInfoLite":
        # TODO: Legacy: remove?
        if (
            particle_dict["hotkey"]
            == "5C4hrfjw9DjXZTzV3MwzrrAr9P1MJhSrvWGWqi1eSuyUpnhM"
        ):
            return ParticleInfoLite._null_particle()
        else:
            particle = ParticleInfoLite(**particle_dict)
            particle.stake = Balance.from_vim(particle.total_stake)
            particle.stake_dict = {
                hk: Balance.from_vim(stake)
                for hk, stake in particle.stake.items()
            }
            particle.total_stake = particle.stake
            particle.rank = particle.rank / U16_MAX
            particle.trust = particle.trust / U16_MAX
            particle.consensus = particle.consensus / U16_MAX
            particle.validator_trust = particle.validator_trust / U16_MAX
            particle.incentive = particle.incentive / U16_MAX
            particle.dividends = particle.dividends / U16_MAX
            particle.emission = particle.emission / VIMPERNIM

            return particle


@dataclass
class ProtonInfo:
    r"""
    Dataclass for proton info.
    """
    block: int
    version: int
    ip: str
    port: int
    ip_type: int

    @classmethod
    def fix_decoded_values(
        cls, proton_info_decoded: Dict
    ) -> "ProtonInfo":
        r"""Returns a ProtonInfo object from a proton_info_decoded dictionary."""
        proton_info_decoded["ip"] = net.int_to_ip(
            int(proton_info_decoded["ip"])
        )

        return cls(**proton_info_decoded)


@dataclass
class DelegateInfo:
    r"""
    Dataclass for delegate info.
    """
    hotkey_ss58: str  # Hotkey of delegate
    total_stake: Balance  # Total stake of the delegate
    nominators: List[
        Tuple[str, Balance]
    ]  # List of nominators of the delegate and their stake
    owner_ss58: str  # Coldkey of owner
    take: float  # Take of the delegate as a percentage
    validator_permits: List[
        int
    ]  # List of cosmos that the delegate is allowed to validate on
    registrations: List[
        int
    ]  # List of cosmos that the delegate is registered on
    return_per_1000: Balance  # Return per 1000 nim of the delegate over a day
    total_daily_return: Balance  # Total daily return of the delegate

    @classmethod
    def fix_decoded_values(cls, decoded: Any) -> "DelegateInfo":
        r"""Fixes the decoded values."""

        return cls(
            hotkey_ss58=ss58_encode(
                decoded["delegate_ss58"], nimble.__ss58_format__
            ),
            owner_ss58=ss58_encode(
                decoded["owner_ss58"], nimble.__ss58_format__
            ),
            take=U16_NORMALIZED_FLOAT(decoded["take"]),
            nominators=[
                (
                    ss58_encode(nom[0], nimble.__ss58_format__),
                    Balance.from_vim(nom[1]),
                )
                for nom in decoded["nominators"]
            ],
            total_stake=Balance.from_vim(
                sum([nom[1] for nom in decoded["nominators"]])
            ),
            validator_permits=decoded["validator_permits"],
            registrations=decoded["registrations"],
            return_per_1000=Balance.from_vim(decoded["return_per_1000"]),
            total_daily_return=Balance.from_vim(decoded["total_daily_return"]),
        )

    @classmethod
    def from_vec_u8(cls, vec_u8: List[int]) -> Optional["DelegateInfo"]:
        r"""Returns a DelegateInfo object from a vec_u8."""
        if len(vec_u8) == 0:
            return None

        decoded = from_scale_encoding(vec_u8, ChainDataType.DelegateInfo)

        if decoded is None:
            return None

        decoded = DelegateInfo.fix_decoded_values(decoded)

        return decoded

    @classmethod
    def list_from_vec_u8(cls, vec_u8: List[int]) -> List["DelegateInfo"]:
        r"""Returns a list of DelegateInfo objects from a vec_u8."""
        decoded = from_scale_encoding(
            vec_u8, ChainDataType.DelegateInfo, is_vec=True
        )

        if decoded is None:
            return []

        decoded = [DelegateInfo.fix_decoded_values(d) for d in decoded]

        return decoded

    @classmethod
    def delegated_list_from_vec_u8(
        cls, vec_u8: List[int]
    ) -> List[Tuple["DelegateInfo", Balance]]:
        r"""Returns a list of Tuples of DelegateInfo objects, and Balance, from a vec_u8.
        This is the list of delegates that the user has delegated to, and the amount of stake delegated.
        """
        decoded = from_scale_encoding(
            vec_u8, ChainDataType.DelegatedInfo, is_vec=True
        )

        if decoded is None:
            return []

        decoded = [
            (DelegateInfo.fix_decoded_values(d), Balance.from_vim(s))
            for d, s in decoded
        ]

        return decoded


@dataclass
class StakeInfo:
    r"""
    Dataclass for stake info.
    """
    hotkey_ss58: str  # Hotkey address
    coldkey_ss58: str  # Coldkey address
    stake: Balance  # Stake for the hotkey-coldkey pair

    @classmethod
    def fix_decoded_values(cls, decoded: Any) -> "StakeInfo":
        r"""Fixes the decoded values."""

        return cls(
            hotkey_ss58=ss58_encode(decoded["hotkey"], nimble.__ss58_format__),
            coldkey_ss58=ss58_encode(
                decoded["coldkey"], nimble.__ss58_format__
            ),
            stake=Balance.from_vim(decoded["stake"]),
        )

    @classmethod
    def from_vec_u8(cls, vec_u8: List[int]) -> Optional["StakeInfo"]:
        r"""Returns a StakeInfo object from a vec_u8."""
        if len(vec_u8) == 0:
            return None

        decoded = from_scale_encoding(vec_u8, ChainDataType.StakeInfo)

        if decoded is None:
            return None

        decoded = StakeInfo.fix_decoded_values(decoded)

        return decoded

    @classmethod
    def list_of_tuple_from_vec_u8(
        cls, vec_u8: List[int]
    ) -> Dict[str, List["StakeInfo"]]:
        r"""Returns a list of StakeInfo objects from a vec_u8."""
        decoded: Optional[
            List[Tuple(str, List[object])]
        ] = from_scale_encoding_using_type_string(
            input=vec_u8, type_string="Vec<(AccountId, Vec<StakeInfo>)>"
        )

        if decoded is None:
            return {}

        stake_map = {
            ss58_encode(
                address=account_id, ss58_format=nimble.__ss58_format__
            ): [StakeInfo.fix_decoded_values(d) for d in stake_info]
            for account_id, stake_info in decoded
        }

        return stake_map

    @classmethod
    def list_from_vec_u8(cls, vec_u8: List[int]) -> List["StakeInfo"]:
        r"""Returns a list of StakeInfo objects from a vec_u8."""
        decoded = from_scale_encoding(
            vec_u8, ChainDataType.StakeInfo, is_vec=True
        )

        if decoded is None:
            return []

        decoded = [StakeInfo.fix_decoded_values(d) for d in decoded]

        return decoded


@dataclass
class CosmosInfo:
    r"""
    Dataclass for cosmos info.
    """
    netuid: int
    rho: int
    kappa: int
    difficulty: int
    immunity_period: int
    max_allowed_validators: int
    min_allowed_weights: int
    max_weight_limit: float
    scaling_law_power: float
    cosmos_n: int
    max_n: int
    blocks_since_epoch: int
    tempo: int
    modality: int
    # netuid -> topk percentile prunning score requirement (u16:MAX normalized.)
    connection_requirements: Dict[str, float]
    emission_value: float
    burn: Balance
    owner_ss58: str

    @classmethod
    def from_vec_u8(cls, vec_u8: List[int]) -> Optional["CosmosInfo"]:
        r"""Returns a CosmosInfo object from a vec_u8."""
        if len(vec_u8) == 0:
            return None

        decoded = from_scale_encoding(vec_u8, ChainDataType.CosmosInfo)

        if decoded is None:
            return None

        return CosmosInfo.fix_decoded_values(decoded)

    @classmethod
    def list_from_vec_u8(cls, vec_u8: List[int]) -> List["CosmosInfo"]:
        r"""Returns a list of CosmosInfo objects from a vec_u8."""
        decoded = from_scale_encoding(
            vec_u8, ChainDataType.CosmosInfo, is_vec=True, is_option=True
        )

        if decoded is None:
            return []

        decoded = [CosmosInfo.fix_decoded_values(d) for d in decoded]

        return decoded

    @classmethod
    def fix_decoded_values(cls, decoded: Dict) -> "CosmosInfo":
        r"""Returns a CosmosInfo object from a decoded CosmosInfo dictionary."""
        return CosmosInfo(
            netuid=decoded["netuid"],
            rho=decoded["rho"],
            kappa=decoded["kappa"],
            difficulty=decoded["difficulty"],
            immunity_period=decoded["immunity_period"],
            max_allowed_validators=decoded["max_allowed_validators"],
            min_allowed_weights=decoded["min_allowed_weights"],
            max_weight_limit=decoded["max_weights_limit"],
            scaling_law_power=decoded["scaling_law_power"],
            cosmos_n=decoded["cosmos_n"],
            max_n=decoded["max_allowed_uids"],
            blocks_since_epoch=decoded["blocks_since_last_step"],
            tempo=decoded["tempo"],
            modality=decoded["network_modality"],
            connection_requirements={
                str(int(netuid)): U16_NORMALIZED_FLOAT(int(req))
                for netuid, req in decoded["network_connect"]
            },
            emission_value=decoded["emission_values"],
            burn=Balance.from_vim(decoded["burn"]),
            owner_ss58=ss58_encode(decoded["owner"], nimble.__ss58_format__),
        )

    def to_parameter_dict(self) -> "torch.nn.ParameterDict":
        r"""Returns a torch tensor of the cosmos info."""
        return torch.nn.ParameterDict(self.__dict__)

    @classmethod
    def from_parameter_dict(
        cls, parameter_dict: "torch.nn.ParameterDict"
    ) -> "CosmosInfo":
        r"""Returns a CosmosInfo object from a torch parameter_dict."""
        return cls(**dict(parameter_dict))


@dataclass
class CosmosHyperparameters:
    r"""
    Dataclass for cosmos hyperparameters.
    """
    rho: int
    kappa: int
    immunity_period: int
    min_allowed_weights: int
    max_weight_limit: float
    tempo: int
    min_difficulty: int
    max_difficulty: int
    weights_version: int
    weights_rate_limit: int
    adjustment_interval: int
    activity_cutoff: int
    registration_allowed: bool
    target_regs_per_interval: int
    min_burn: int
    max_burn: int
    bonds_moving_avg: int
    max_regs_per_block: int
    serving_rate_limit: int
    max_validators: int

    @classmethod
    def from_vec_u8(
        cls, vec_u8: List[int]
    ) -> Optional["CosmosHyperparameters"]:
        r"""Returns a CosmosHyperparameters object from a vec_u8."""
        if len(vec_u8) == 0:
            return None

        decoded = from_scale_encoding(
            vec_u8, ChainDataType.CosmosHyperparameters
        )

        if decoded is None:
            return None

        return CosmosHyperparameters.fix_decoded_values(decoded)

    @classmethod
    def list_from_vec_u8(
        cls, vec_u8: List[int]
    ) -> List["CosmosHyperparameters"]:
        r"""Returns a list of CosmosHyperparameters objects from a vec_u8."""
        decoded = from_scale_encoding(
            vec_u8,
            ChainDataType.CosmosHyperparameters,
            is_vec=True,
            is_option=True,
        )

        if decoded is None:
            return []

        decoded = [CosmosHyperparameters.fix_decoded_values(d) for d in decoded]

        return decoded

    @classmethod
    def fix_decoded_values(cls, decoded: Dict) -> "CosmosHyperparameters":
        r"""Returns a CosmosInfo object from a decoded CosmosInfo dictionary."""
        return CosmosHyperparameters(
            rho=decoded["rho"],
            kappa=decoded["kappa"],
            immunity_period=decoded["immunity_period"],
            min_allowed_weights=decoded["min_allowed_weights"],
            max_weight_limit=decoded["max_weights_limit"],
            tempo=decoded["tempo"],
            min_difficulty=decoded["min_difficulty"],
            max_difficulty=decoded["max_difficulty"],
            weights_version=decoded["weights_version"],
            weights_rate_limit=decoded["weights_rate_limit"],
            adjustment_interval=decoded["adjustment_interval"],
            activity_cutoff=decoded["activity_cutoff"],
            registration_allowed=decoded["registration_allowed"],
            target_regs_per_interval=decoded["target_regs_per_interval"],
            min_burn=decoded["min_burn"],
            max_burn=decoded["max_burn"],
            bonds_moving_avg=decoded["bonds_moving_avg"],
            max_regs_per_block=decoded["max_regs_per_block"],
            max_validators=decoded["max_validators"],
            serving_rate_limit=decoded["serving_rate_limit"],
        )

    def to_parameter_dict(self) -> "torch.nn.ParameterDict":
        r"""Returns a torch tensor of the cosmos hyperparameters."""
        return torch.nn.ParameterDict(self.__dict__)

    @classmethod
    def from_parameter_dict(
        cls, parameter_dict: "torch.nn.ParameterDict"
    ) -> "CosmosInfo":
        r"""Returns a CosmosHyperparameters object from a torch parameter_dict."""
        return cls(**dict(parameter_dict))


@dataclass
class IPInfo:
    r"""
    Dataclass for associated IP Info.
    """
    ip: str
    ip_type: int
    protocol: int

    def encode(self) -> Dict[str, Any]:
        r"""Returns a dictionary of the IPInfo object that can be encoded."""
        return {
            "ip": net.ip_to_int(
                self.ip
            ),  # IP type and protocol are encoded together as a u8
            "ip_type_and_protocol": ((self.ip_type << 4) + self.protocol)
            & 0xFF,
        }

    @classmethod
    def from_vec_u8(cls, vec_u8: List[int]) -> Optional["IPInfo"]:
        r"""Returns a IPInfo object from a vec_u8."""
        if len(vec_u8) == 0:
            return None

        decoded = from_scale_encoding(vec_u8, ChainDataType.IPInfo)

        if decoded is None:
            return None

        return IPInfo.fix_decoded_values(decoded)

    @classmethod
    def list_from_vec_u8(cls, vec_u8: List[int]) -> List["IPInfo"]:
        r"""Returns a list of IPInfo objects from a vec_u8."""
        decoded = from_scale_encoding(vec_u8, ChainDataType.IPInfo, is_vec=True)

        if decoded is None:
            return []

        decoded = [IPInfo.fix_decoded_values(d) for d in decoded]

        return decoded

    @classmethod
    def fix_decoded_values(cls, decoded: Dict) -> "IPInfo":
        r"""Returns a CosmosInfo object from a decoded IPInfo dictionary."""
        return IPInfo(
            ip=nimble.utils.networking.int_to_ip(decoded["ip"]),
            ip_type=decoded["ip_type_and_protocol"] >> 4,
            protocol=decoded["ip_type_and_protocol"] & 0xF,
        )

    def to_parameter_dict(self) -> "torch.nn.ParameterDict":
        r"""Returns a torch tensor of the cosmos info."""
        return torch.nn.ParameterDict(self.__dict__)

    @classmethod
    def from_parameter_dict(
        cls, parameter_dict: "torch.nn.ParameterDict"
    ) -> "IPInfo":
        r"""Returns a IPInfo object from a torch parameter_dict."""
        return cls(**dict(parameter_dict))


# Senate / Proposal data


class ProposalVoteData(TypedDict):
    index: int
    threshold: int
    ayes: List[str]
    nays: List[str]
    end: int


ProposalCallData = GenericCall



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import nimble

import torch
import time
from rich.prompt import Confirm
from typing import List, Union, Optional, Tuple
from nimble.utils.registration import POWSolution, create_pow


def register_extrinsic(
    nbnetwork: "nimble.nbnetwork",
    wallet: "nimble.wallet",
    netuid: int,
    wait_for_inclusion: bool = False,
    wait_for_finalization: bool = True,
    prompt: bool = False,
    max_allowed_attempts: int = 3,
    output_in_place: bool = True,
    cuda: bool = False,
    dev_id: Union[List[int], int] = 0,
    tpb: int = 256,
    num_processes: Optional[int] = None,
    update_interval: Optional[int] = None,
    log_verbose: bool = False,
) -> bool:
    r"""Registers the wallet to chain.
    Args:
        wallet (nimble.wallet):
            nimble wallet object.
        netuid (int):
            The netuid of the subnet to register on.
        wait_for_inclusion (bool):
            If set, waits for the extrinsic to enter a block before returning true,
            or returns false if the extrinsic fails to enter the block within the timeout.
        wait_for_finalization (bool):
            If set, waits for the extrinsic to be finalized on the chain before returning true,
            or returns false if the extrinsic fails to be finalized within the timeout.
        prompt (bool):
            If true, the call waits for confirmation from the user before proceeding.
        max_allowed_attempts (int):
            Maximum number of attempts to register the wallet.
        cuda (bool):
            If true, the wallet should be registered using CUDA device(s).
        dev_id (Union[List[int], int]):
            The CUDA device id to use, or a list of device ids.
        tpb (int):
            The number of threads per block (CUDA).
        num_processes (int):
            The number of processes to use to register.
        update_interval (int):
            The number of nonces to solve between updates.
        log_verbose (bool):
            If true, the registration process will log more information.
    Returns:
        success (bool):
            flag is true if extrinsic was finalized or uncluded in the block.
            If we did not wait for finalization / inclusion, the response is true.
    """
    if not nbnetwork.subnet_exists(netuid):
        nimble.__console__.print(
            ":cross_mark: [red]Failed[/red]: error: [bold white]subnet:{}[/bold white] does not exist.".format(
                netuid
            )
        )
        return False

    with nimble.__console__.status(
        f":satellite: Checking Account on [bold]subnet:{netuid}[/bold]..."
    ):
        neuron = nbnetwork.get_neuron_for_pubkey_and_subnet(
            wallet.hotkey.ss58_address, netuid=netuid
        )
        if not neuron.is_null:
            nimble.logging.debug(
                f"Wallet {wallet} is already registered on {neuron.netuid} with {neuron.uid}"
            )
            return True

    if prompt:
        if not Confirm.ask(
            "Continue Registration?\n  hotkey:     [bold white]{}[/bold white]\n  coldkey:    [bold white]{}[/bold white]\n  network:    [bold white]{}[/bold white]".format(
                wallet.hotkey.ss58_address,
                wallet.coldkeypub.ss58_address,
                nbnetwork.network,
            )
        ):
            return False

    # Attempt rolling registration.
    attempts = 1
    while True:
        nimble.__console__.print(
            ":satellite: Registering...({}/{})".format(
                attempts, max_allowed_attempts
            )
        )
        # Solve latest POW.
        if cuda:
            if not torch.cuda.is_available():
                if prompt:
                    nimble.__console__.error("CUDA is not available.")
                return False
            pow_result: Optional[POWSolution] = create_pow(
                nbnetwork,
                wallet,
                netuid,
                output_in_place,
                cuda=cuda,
                dev_id=dev_id,
                tpb=tpb,
                num_processes=num_processes,
                update_interval=update_interval,
                log_verbose=log_verbose,
            )
        else:
            pow_result: Optional[POWSolution] = create_pow(
                nbnetwork,
                wallet,
                netuid,
                output_in_place,
                cuda=cuda,
                num_processes=num_processes,
                update_interval=update_interval,
                log_verbose=log_verbose,
            )

        # pow failed
        if not pow_result:
            # might be registered already on this subnet
            is_registered = nbnetwork.is_hotkey_registered(
                netuid=netuid, hotkey_ss58=wallet.hotkey.ss58_address
            )
            if is_registered:
                nimble.__console__.print(
                    f":white_heavy_check_mark: [green]Already registered on netuid:{netuid}[/green]"
                )
                return True

        # pow successful, proceed to submit pow to chain for registration
        else:
            with nimble.__console__.status(":satellite: Submitting POW..."):
                # check if pow result is still valid
                while not pow_result.is_stale(nbnetwork=nbnetwork):
                    result: Tuple[
                        bool, Optional[str]
                    ] = nbnetwork._do_pow_register(
                        netuid=netuid,
                        wallet=wallet,
                        pow_result=pow_result,
                        wait_for_inclusion=wait_for_inclusion,
                        wait_for_finalization=wait_for_finalization,
                    )
                    success, err_msg = result

                    if success != True or success == False:
                        if "key is already registered" in err_msg:
                            # Error meant that the key is already registered.
                            nimble.__console__.print(
                                f":white_heavy_check_mark: [green]Already Registered on [bold]subnet:{netuid}[/bold][/green]"
                            )
                            return True

                        nimble.__console__.print(
                            ":cross_mark: [red]Failed[/red]: error:{}".format(
                                err_msg
                            )
                        )
                        time.sleep(0.5)

                    # Successful registration, final check for neuron and pubkey
                    else:
                        nimble.__console__.print(
                            ":satellite: Checking Balance..."
                        )
                        is_registered = nbnetwork.is_hotkey_registered(
                            netuid=netuid,
                            hotkey_ss58=wallet.hotkey.ss58_address,
                        )
                        if is_registered:
                            nimble.__console__.print(
                                ":white_heavy_check_mark: [green]Registered[/green]"
                            )
                            return True
                        else:
                            # neuron not found, try again
                            nimble.__console__.print(
                                ":cross_mark: [red]Unknown error. Neuron not found.[/red]"
                            )
                            continue
                else:
                    # Exited loop because pow is no longer valid.
                    nimble.__console__.print("[red]POW is stale.[/red]")
                    # Try again.
                    continue

        if attempts < max_allowed_attempts:
            # Failed registration, retry pow
            attempts += 1
            nimble.__console__.print(
                ":satellite: Failed registration, retrying pow ...({}/{})".format(
                    attempts, max_allowed_attempts
                )
            )
        else:
            # Failed to register after max attempts.
            nimble.__console__.print("[red]No more attempts.[/red]")
            return False


def burned_register_extrinsic(
    nbnetwork: "nimble.nbnetwork",
    wallet: "nimble.wallet",
    netuid: int,
    wait_for_inclusion: bool = False,
    wait_for_finalization: bool = True,
    prompt: bool = False,
) -> bool:
    r"""Registers the wallet to chain by recycling NIM.
    Args:
        wallet (nimble.wallet):
            nimble wallet object.
        netuid (int):
            The netuid of the subnet to register on.
        wait_for_inclusion (bool):
            If set, waits for the extrinsic to enter a block before returning true,
            or returns false if the extrinsic fails to enter the block within the timeout.
        wait_for_finalization (bool):
            If set, waits for the extrinsic to be finalized on the chain before returning true,
            or returns false if the extrinsic fails to be finalized within the timeout.
        prompt (bool):
            If true, the call waits for confirmation from the user before proceeding.
    Returns:
        success (bool):
            flag is true if extrinsic was finalized or uncluded in the block.
            If we did not wait for finalization / inclusion, the response is true.
    """
    if not nbnetwork.subnet_exists(netuid):
        nimble.__console__.print(
            ":cross_mark: [red]Failed[/red]: error: [bold white]subnet:{}[/bold white] does not exist.".format(
                netuid
            )
        )
        return False

    wallet.coldkey  # unlock coldkey
    with nimble.__console__.status(
        f":satellite: Checking Account on [bold]subnet:{netuid}[/bold]..."
    ):
        neuron = nbnetwork.get_neuron_for_pubkey_and_subnet(
            wallet.hotkey.ss58_address, netuid=netuid
        )

        old_balance = nbnetwork.get_balance(wallet.coldkeypub.ss58_address)

        burn_amount = nbnetwork.burn(netuid=netuid)
        if not neuron.is_null:
            nimble.__console__.print(
                ":white_heavy_check_mark: [green]Already Registered[/green]:\n"
                "uid: [bold white]{}[/bold white]\n"
                "netuid: [bold white]{}[/bold white]\n"
                "hotkey: [bold white]{}[/bold white]\n"
                "coldkey: [bold white]{}[/bold white]".format(
                    neuron.uid, neuron.netuid, neuron.hotkey, neuron.coldkey
                )
            )
            return True

    if prompt:
        # Prompt user for confirmation.
        if not Confirm.ask(
            f"Recycle {burn_amount} to register on subnet:{netuid}?"
        ):
            return False

    with nimble.__console__.status(
        ":satellite: Recycling NIM for Registration..."
    ):
        success, err_msg = nbnetwork._do_burned_register(
            netuid=netuid,
            wallet=wallet,
            wait_for_inclusion=wait_for_inclusion,
            wait_for_finalization=wait_for_finalization,
        )

        if success != True or success == False:
            nimble.__console__.print(
                ":cross_mark: [red]Failed[/red]: error:{}".format(err_msg)
            )
            time.sleep(0.5)

        # Successful registration, final check for neuron and pubkey
        else:
            nimble.__console__.print(":satellite: Checking Balance...")
            block = nbnetwork.get_current_block()
            new_balance = nbnetwork.get_balance(
                wallet.coldkeypub.ss58_address, block=block
            )

            nimble.__console__.print(
                "Balance:\n  [blue]{}[/blue] :arrow_right: [green]{}[/green]".format(
                    old_balance, new_balance
                )
            )
            is_registered = nbnetwork.is_hotkey_registered(
                netuid=netuid, hotkey_ss58=wallet.hotkey.ss58_address
            )
            if is_registered:
                nimble.__console__.print(
                    ":white_heavy_check_mark: [green]Registered[/green]"
                )
                return True
            else:
                # neuron not found, try again
                nimble.__console__.print(
                    ":cross_mark: [red]Unknown error. Neuron not found.[/red]"
                )


class MaxSuccessException(Exception):
    pass


class MaxAttemptsException(Exception):
    pass